"""This module describes all errors which can be reported by Darglint.

Errors can be anything from being unable to parse a docstring,
to having docstring arguments out of sync with the function/method
definition.

Groups of errors:

    000 Style
    100 Args
    200 Returns
    300 Yields
    400 Raises
    500 Variables

These errors are based on the errors and warnings from the
pycodestyle package.  Interface, here, describes incorrect or incomplete
documentation.  Style, here, means errors in documentation style.
So, for instance, missing a parameter in the documentation would be
DAR101.

"""
import ast  # noqa: F401
from typing import (
    Tuple,
    Union,
)


class DarglintError(BaseException):
    """The base error class for any darglint error."""

    # The shortest error message possible.  Should use abbreviated
    # symbols.  Can be combined with the general message to give
    # a more in-depth message.
    terse_message = None  # type: str

    # General messages should be general in nature. (They are used
    # at the lowest verbosity setting.)  They should be about
    # the nature of the error, and not about this particular instance.
    # Should not end in any punctuation (and should allow lists of
    # instances -- the terse message -- after it.)
    general_message = None  # type: str

    # A unique human-readable identifier for this type of error.
    # See the description of error code groups above.
    error_code = None  # type: str

    # The first and last line numbers where the error occurs.
    line_numbers = None  # type: Tuple[int, int]

    # A description of the error itself.
    description = None  # type: str

    def message(self, verbosity=1):
        # type: (int) -> str
        """Get the message for this error, according to the verbosity.

        Args:
            verbosity: An integer in the set {1,2}, where 1 is a more
                terse message, and 2 includes a general description.

        Raises:
            Exception: If the verbosity level is not recognized.

        Returns:
            An error message.

        """
        if verbosity == 1:
            return '{}'.format(self.terse_message)
        elif verbosity == 2:
            return '{}: {}'.format(
                self.general_message,
                self.terse_message,
            )
        else:
            raise Exception('Unrecognized verbosity setting, {}.'.format(
                verbosity))

    def __init__(self, function, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], Tuple[int, int]) -> None
        """Create a new exception with a message and line number.

        Raises:
            NotImplementedError: If the required methods in this
                base class have not been implemented.

        Args:
            function: An ast node for the function.
            line_numbers: The line numbers where this error occurs.

        """
        self.function = function

        if line_numbers:
            self.line_numbers = line_numbers

        # The abstract base class syntax was too verbose for this,
        # and not really justified by the size of the module.
        if (self.terse_message is None
                or self.general_message is None
                or self.error_code is None):
            raise NotImplementedError


class PythonSyntaxError(DarglintError):
    """Describes a syntax error in parsing the python module.."""

    error_code = 'DAR000'
    description = (
        'Failed to parse file due to syntax error.'
    )

    def __init__(self, source):
        # type: (SyntaxError) -> None  # noqa: E501
        """Instantiate the error's message.

        Args:
            source: The syntax error which was raised by ast.

        """
        self.general_message = 'Python syntax error'
        self.terse_message = 's {}'.format(source)
        self.line_numbers = (source.lineno or 0, source.lineno or 0)


class GenericSyntaxError(DarglintError):
    """Describes that something went wrong in parsing the docstring."""

    error_code = 'DAR001'
    description = (
        'The docstring was not parsed correctly due to a syntax error.'
    )

    def __init__(self, function, message, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], str, Tuple[int, int]) -> None  # noqa: E501
        """Instantiate the error's message.

        Args:
            function: An ast node for the function.
            message: The parser error's message.
            line_numbers: The line numbers where this error occurs.

        """
        self.general_message = 'Syntax error'
        self.terse_message = 's {}'.format(message)

        super(GenericSyntaxError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class EmptyDescriptionError(DarglintError):
    """Describes when an argument/exception lacks a description."""

    error_code = 'DAR002'
    description = 'An argument/exception lacks a description'

    def __init__(self, function, message, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], str, Tuple[int, int]) -> None  # noqa: E501
        """Instantiate the error's message.

        Args:
            function: An ast node for the function.
            message: The parser error's message.
            line_numbers: The line numbers where this error occurs.
                Unused.

        """
        self.general_message = 'Empty description'
        self.terse_message = 'e {}'.format(message)

        super(EmptyDescriptionError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class IndentError(DarglintError):
    """Describes when a line is under-indented or over-indented.
    """

    error_code = 'DAR003'
    description = 'A line is under-indented or over-indented.'

    def __init__(self, function, line_numbers=None):
        # type: (ast.FunctionDef, Tuple[int, int]) -> None
        """Instantiate the eror's message.

        Args:
            function: An ast node for the function.
            line_numbers: The line numbers where this error occurs.

        """
        self.general_message = 'Incorrect indentation'
        self.terse_message = '~<'

        super(IndentError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class ExcessNewlineError(DarglintError):
    """Describes when a docstring has an extra newline where it shouldn't."""

    error_code = 'DAR004'
    description = (
        'The docstring contains an extra newline where it shouldn\'t.'
    )

    def __init__(self, function, line_numbers=None):
        # type: (ast.FunctionDef, Tuple[int, int]) -> None
        self.general_message = 'Excess newline.'
        self.terse_message = '+>'
        super(ExcessNewlineError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class EmptyTypeError(DarglintError):
    """Describes when an item has parentheses, but no type."""

    error_code = 'DAR005'
    description = (
        'The item contains a type section (parentheses), but no type.'
    )

    def __init__(self, function, message, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], str, Tuple[int, int]) -> None  # noqa: E501
        """Instantiate the error's message.

        Args:
            function: An ast node for the function.
            message: The parser error's message.
            line_numbers: The line numbers where this error occurs.
                Unused.

        """
        self.general_message = 'Empty type'
        self.terse_message = 'e {}'.format(message)

        super(EmptyTypeError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class MissingParameterError(DarglintError):
    """Describes when a docstring is missing a parameter in the definition."""

    error_code = 'DAR101'
    description = 'The docstring is missing a parameter in the definition.'

    def __init__(self, function, name, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], str, Tuple[int, int]) -> None
        """Instantiate the error's message.

        Args:
            function: An ast node for the function.
            name: The name of the argument that is missing.
            line_numbers: The line numbers where this error occurs.

        """
        self.general_message = 'Missing parameter(s) in Docstring'
        self.terse_message = '- {}'.format(name)
        super(MissingParameterError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class ExcessParameterError(DarglintError):
    """Describes when a docstring contains a parameter not in function."""

    error_code = 'DAR102'
    description = 'The docstring contains a parameter not in function.'

    def __init__(self, function, name, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], str, Tuple[int, int]) -> None
        """Instantiate the error's message.

        Args:
            function: An ast node for the function.
            name: The name of the argument that is excess.
            line_numbers: The line numbers where this error occurs.

        """
        self.general_message = 'Excess parameter(s) in Docstring'
        self.terse_message = '+ {}'.format(name)
        super(ExcessParameterError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class ParameterTypeMismatchError(DarglintError):
    """Describes when a docstring parameter type doesn't match function."""

    error_code = 'DAR103'
    description = 'The docstring parameter type doesn\'t match function.'

    def __init__(self, function, name, expected, actual, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], str, str, str, Tuple[int, int]) -> None
        """Instantiate the error's message.

        Args:
            function: An ast node for the function.
            name: The name of the parameter.
            expected: The type defined in the function.
            actual: The type described in the docstring.
            line_numbers: The line numbers where this error occurs.

        """
        self.general_message = 'Parameter type mismatch'
        self.terse_message = ' ~{}: expected {} but was {}'.format(
            name,
            expected,
            actual,
        )
        self.name = name
        self.expected = expected
        self.actual = actual
        super(ParameterTypeMismatchError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class ParameterTypeMissingError(DarglintError):
    """Describes when a argument definition has no type specified"""

    error_code = 'DAR104'
    description = 'The docstring parameter type is not given.'

    def __init__(self, function, name, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], str, Tuple[int, int]) -> None
        """Instantiate the error's message.

        Args:
            function: An ast node for the function.
            name: The name of the parameter.
            line_numbers: The line numbers where this error occurs.

        """
        self.general_message = 'Parameter type missing'
        self.terse_message = ' ~{}: type was not specified'.format(
            name
        )
        self.name = name
        super(ParameterTypeMissingError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class ParameterMalformedError(DarglintError):
    """Describes when an argument type is malformed."""

    error_code = 'DAR105'
    description = (
        'The docstring parameter type is malformed. '
        'Expected parameter type to match syntax from PEP484. '
        'If your parameter contains parentheses, you may want '
        'to switch them to brackets.  E.g. `Union[str, int]`.'
    )

    def __init__(self, function, name, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], str, Tuple[int, int]) -> None
        self.general_message = 'Parameter type malformed.'
        self.terse_message = ' #{}: type malformed'.format(
            name
        )
        self.name = name
        super(ParameterMalformedError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class MissingReturnError(DarglintError):
    """Describes when a docstring is missing a return from definition."""

    error_code = 'DAR201'
    description = 'The docstring is missing a return from definition.'

    def __init__(self, function, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], Tuple[int, int]) -> None
        """Instantiate the error's message.

        Args:
            function: An ast node for the function.
            line_numbers: The line numbers where this error occurs.

        """
        self.general_message = 'Missing "Returns" in Docstring'
        self.terse_message = '- return'

        super(MissingReturnError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class ExcessReturnError(DarglintError):
    """Describes when a docstring has a return not in definition."""

    error_code = 'DAR202'
    description = 'The docstring has a return not in definition.'

    def __init__(self, function, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], Tuple[int, int]) -> None
        """Instantiate the error's message.

        Args:
            function: An ast node for the function.
            line_numbers: The line numbers where this error occurs.

        """
        self.general_message = 'Excess "Returns" in Docstring'
        self.terse_message = '+ return'

        super(ExcessReturnError, self).__init__(
            function,
            line_numbers=line_numbers,
        )

class ReturnTypeMismatchError(DarglintError):
    """Describes when a docstring parameter type doesn't match function."""

    error_code = 'DAR203'
    description = 'The docstring parameter type doesn\'t match function.'

    def __init__(self, function, expected, actual, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], str, str, Tuple[int, int]) -> None
        """Instantiate the error's message.

        Args:
            function: An ast node for the function.
            expected: The type defined in the function.
            actual: The type described in the docstring.
            line_numbers: The line numbers where this error occurs.

        """
        self.general_message = 'Return type mismatch'
        self.terse_message = ' ~Return: expected {} but was {}'.format(
            expected,
            actual,
        )
        self.expected = expected
        self.actual = actual
        super(ReturnTypeMismatchError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class MissingYieldError(DarglintError):
    """Describes when a docstring is missing a yield present in definition."""

    error_code = 'DAR301'
    description = 'The docstring is missing a yield present in definition.'

    def __init__(self, function, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], Tuple[int, int]) -> None
        """Instantiate the error's message.

        Args:
            function: An ast node for the function.
            line_numbers: The line numbers where this error occurs.

        """
        self.general_message = 'Missing "Yields" in Docstring'
        self.terse_message = '- yield'

        super(MissingYieldError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class ExcessYieldError(DarglintError):
    """Describes when a docstring has a yield not in definition."""

    error_code = 'DAR302'
    description = 'The docstring has a yield not in definition.'

    def __init__(self, function, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], Tuple[int, int]) -> None
        """Instantiate the error's message.

        Args:
            function: An ast node for the function.
            line_numbers: The line numbers where this error occurs.

        """
        self.general_message = 'Excess "Yields" in Docstring'
        self.terse_message = '+ yield'

        super(ExcessYieldError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class MissingRaiseError(DarglintError):
    """Describes when a docstring is missing an exception raised."""

    error_code = 'DAR401'
    description = 'The docstring is missing an exception raised.'

    def __init__(self, function, name, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], str, Tuple[int, int]) -> None
        """Instantiate the error's message.

        Args:
            function: An ast node for the function.
            name: The name of the exception that is missing.
            line_numbers: The line numbers where this error occurs.

        """
        self.general_message = 'Missing exception(s) in Raises section'
        self.terse_message = '-r {}'.format(name)
        self.name = name
        super(MissingRaiseError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class ExcessRaiseError(DarglintError):
    """Describes when a docstring describes an exception not explicitly raised.

    This error should not be included by default.  We assume that the user
    knows when an underlying function is likely to raise an error.  Of course,
    we should provide the option, if the user wants to be explicit.  (And
    catch and reraise an exception.)

    """

    error_code = 'DAR402'
    description = 'The docstring describes an exception not explicitly raised.'

    def __init__(self, function, name, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], str, Tuple[int, int]) -> None
        """Instantiate the error's message.

        Args:
            function: An ast node for the function.
            name: The name of the exception that is surplus.
            line_numbers: The line numbers where this error occurs.

        """
        self.general_message = 'Excess exception(s) in Raises section'
        self.terse_message = '+r {}'.format(name)
        self.name = name
        super(ExcessRaiseError, self).__init__(
            function,
            line_numbers=line_numbers,
        )


class ExcessVariableError(DarglintError):
    """Describes when a docstring describes a variable which is not defined."""

    error_code = 'DAR501'
    description = 'The docstring describes a variable which is not defined.'

    def __init__(self, function, name, line_numbers=None):
        # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], str, Tuple[int, int]) -> None
        """Instantiate the error's message.

        Args:
            function: The ast node for the function.
            name: The name of the variable which is in excess.
            line_numbers: The first and last line numbers where this
                error occurs.

        """
        self.general_message = 'Excess variable description.'
        self.terse_message = '+v {}'.format(name)
        self.name = name
        super(ExcessVariableError, self).__init__(
            function,
            line_numbers=line_numbers,
        )



"""Defines a function for lexing a comment, `lex`."""

from typing import (
    Iterator,
    List,
    Optional,
)
from .custom_assert import Assert
from .peaker import Peaker
from .token import Token, TokenType
from .config import (
    get_config,
)

# These convenience functions take an optional string
# because the peaker could return None when at the end
# of the stream.


def _is_space(char):
    # type: (Optional[str]) -> bool
    return char == ' '


def _is_newline(char):
    # type: (Optional[str]) -> bool
    return char == '\n'


def _is_colon(char):
    # type: (Optional[str]) -> bool
    return char == ':'


def _is_hash(char):
    # type: (Optional[str]) -> bool
    return char == '#'


def _is_separator(char):
    # type: (Optional[str]) -> bool
    """Check whether if `char` is a separator other than newline or space.

    Args:
        char: The character to check.

    Returns:
        true if `char` is a separator other than newline or space.

    """
    if char is None:
        return False
    return char.isspace() and not (_is_space(char) or _is_newline(char))


def _is_lparen(char):
    # type: (Optional[str]) -> bool
    return char == '('


def _is_rparen(char):
    # type: (Optional[str]) -> bool
    return char == ')'


def _is_hyphen(char):
    # type: (Optional[str]) -> bool
    return char == '-'


def _is_word(char):
    # type: (str) -> bool
    return not any([
        _is_space(char),
        _is_newline(char),
        _is_colon(char),
        _is_separator(char),
        _is_hash(char),
        _is_lparen(char),
        _is_rparen(char),
    ])


def lex(program):
    # type: (str) -> Iterator[Token]
    """Create a stream of tokens from the string.

    Args:
        program: The program to lex, as a string.

    Yields:
        Tokens lexed from the string.

    """
    extra = ''  # Extra characters which are pulled but unused from a check.
    peaker = Peaker((x for x in program or []))  # the stream
    line_number = 0

    # Set the amount of spaces which count as an indent.
    config = get_config()

    while peaker.has_next():
        # Each of the following conditions must move the stream
        # forward and -- excepting separators -- yield a token.
        if _is_space(peaker.peak()):
            spaces = ''.join(peaker.take_while(_is_space))
            for _ in range(len(spaces) // config.indentation):
                yield Token(' ' * 4, TokenType.INDENT, line_number)
        elif _is_newline(peaker.peak()):
            value = peaker.next()
            yield Token(value, TokenType.NEWLINE, line_number)
            line_number += 1
        elif _is_colon(peaker.peak()):
            value = peaker.next()
            yield Token(value, TokenType.COLON, line_number)
        elif _is_separator(peaker.peak()):
            peaker.take_while(_is_separator)
        elif _is_hash(peaker.peak()):
            value = peaker.next()
            yield Token(value, TokenType.HASH, line_number)
        elif _is_lparen(peaker.peak()):
            value = peaker.next()
            yield Token(value, TokenType.LPAREN, line_number)
        elif _is_rparen(peaker.peak()):
            value = peaker.next()
            yield Token(value, TokenType.RPAREN, line_number)
        elif _is_hyphen(peaker.peak()):
            value = ''.join(peaker.take_while(_is_word))
            if value.count('-') == len(value):
                yield Token(value, TokenType.HEADER, line_number)
            else:
                yield Token(value, TokenType.WORD, line_number)
        else:
            value = ''.join(peaker.take_while(_is_word))
            if extra != '':
                value = extra + value
                extra = ''
            Assert(
                len(value) > 0,
                "There should be non-special characters.",
            )
            yield Token(value, TokenType.WORD, line_number)


KEYWORDS = {
    'Args': TokenType.ARGUMENTS,
    'Arguments': TokenType.ARGUMENTS,
    'Yields': TokenType.YIELDS,
    'Raises': TokenType.RAISES,
    'Returns': TokenType.RETURNS,
    'noqa': TokenType.NOQA,
    'param': TokenType.ARGUMENTS,
    'parameter': TokenType.ARGUMENTS,
    'Parameters': TokenType.ARGUMENTS,
    'arg': TokenType.ARGUMENTS,
    'argument': TokenType.ARGUMENTS,
    'key': TokenType.VARIABLES,
    'keyword': TokenType.VARIABLES,
    'var': TokenType.VARIABLES,
    'ivar': TokenType.VARIABLES,
    'cvar': TokenType.VARIABLES,
    'type': TokenType.ARGUMENT_TYPE,
    'vartype': TokenType.VARIABLE_TYPE,
    'raises': TokenType.RAISES,
    'yield': TokenType.YIELDS,
    'yields': TokenType.YIELDS,
    'ytype': TokenType.YIELD_TYPE,
    'return': TokenType.RETURNS,
    'returns': TokenType.RETURNS,
    'rtype': TokenType.RETURN_TYPE,
    'Other': TokenType.OTHER,
    'Receives': TokenType.RECEIVES,
    'Warns': TokenType.WARNS,
    'Warnings': TokenType.WARNS,
    'See': TokenType.SEE,
    'Also': TokenType.ALSO,
    'Notes': TokenType.NOTES,
    'Examples': TokenType.EXAMPLES,
    'References': TokenType.REFERENCES,
}


def condense(tokens):
    # type: (Iterator[Token]) -> List[Token]
    """Condense the stream of tokens into a list consumable by CYK.

    This servers two purposes:

    1. It minimizes the lookup table used in the CYK algorithm.
       (The CYK algorithm is a dynamic algorithm, with one dimension
       in the two-dimension lookup table being determined by the number
       of tokens.)

    2. It applies more discriminate token types to the tokens identified
       by lex.  Eventually, this will be moved into the lex function.

    Args:
        tokens: The stream of tokens from the lex function.

    Returns:
        A List of tokens which have been condensed into as small a
        representation as possible.

    """
    ret = list()  # type: List[Token]
    try:
        curr = next(tokens)
    except StopIteration:
        return ret

    if curr.value in KEYWORDS:
        curr.token_type = KEYWORDS[curr.value]

    encountered_noqa = False

    for token in tokens:
        if token.token_type == TokenType.WORD and token.value in KEYWORDS:
            ret.append(curr)
            if token.value == 'noqa':
                encountered_noqa = True
            curr = Token(
                token.value,
                KEYWORDS[token.value],
                token.line_number,
            )
        elif token.token_type == TokenType.WORD:
            if curr.token_type == TokenType.WORD and not encountered_noqa:
                curr.value += ' {}'.format(token.value)
            else:
                ret.append(curr)
                curr = token
        elif token.token_type == TokenType.NEWLINE:
            ret.append(curr)
            curr = token
            encountered_noqa = False
        else:
            ret.append(curr)
            curr = token

    ret.append(curr)

    return ret



import ast
from typing import (
    Any,
    Dict,
    List,
    Union,
)
from ..config import (
    get_logger,
)
from ..custom_assert import (
    Assert,
)


class YieldVisitor(ast.NodeVisitor):
    """A visitor which checks for *returns* nodes."""

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None

        # TODO: https://github.com/python/mypy/issues/4001
        super(YieldVisitor, self).__init__(*args, **kwargs)  # type: ignore

        # A list of the return nodes encountered.
        self.yields = list()  # type: List[Union[ast.Yield, ast.YieldFrom]]

    def visit_Yield(self, node):
        # type: (ast.Yield) -> ast.AST
        self.yields.append(node)
        return self.generic_visit(node)

    def visit_YieldFrom(self, node):
        # type: (ast.YieldFrom) -> ast.AST
        self.yields.append(node)
        return self.generic_visit(node)



import inspect
from typing import (
    List,
    Optional,
    Callable,
    Union,
    Dict,
)
from functools import (
    reduce,
)
from ..token import (
    Token,
    TokenType,
)
from ..node import (
    CykNode,
)
from ..custom_assert import Assert
from .cyk import (
    parse as cyk_parse,
)
from .grammars.numpy_arguments_section import (
    ArgumentsGrammar,
)
from .grammars.numpy_other_arguments_section import (
    OtherArgumentsGrammar,
)
from .grammars.numpy_raises_section import (
    RaisesGrammar,
)
from .grammars.numpy_receives_section import (
    ReceivesGrammar,
)
from .grammars.numpy_returns_section import (
    ReturnsGrammar,
)
from .grammars.numpy_short_description import (
    ShortDescriptionGrammar,
)
from .grammars.numpy_warns_section import (
    WarnsGrammar,
)
from .grammars.numpy_yields_section import (
    YieldsGrammar,
)
from .grammar import (
    BaseGrammar,
)
from .combinator import (
    parser_combinator,
)
from ..token import (
    KEYWORDS,
)
from .long_description import (
    parse as long_description_parse,
)


def top_parse(tokens):
    # type: (List[Token]) -> List[List[Token]]
    """Split the docstring into sections.

    Each section will be parsed individually, according
    to the combinator.

    Args:
        tokens: The tokens representing the entire docstring.

    Returns:
        The docstring, split into sections.

    """
    def at_section_boundary(i):
        # type: (int) -> int
        """Return the number of tokens which start this new section.

        Args:
            i: The current pointer in the token list.

        Returns:
            The number of tokens which start this new section.
            If 0, then this is not a new section.

        """
        if i >= len(tokens):
            return 0
        if tokens[i].token_type == TokenType.OTHER:
            if (
                i + 3 < len(tokens) and
                tokens[i + 1].token_type == TokenType.ARGUMENTS and
                tokens[i + 2].token_type == TokenType.NEWLINE and
                tokens[i + 3].token_type == TokenType.HEADER
            ):
                return 4
            else:
                return 0
        elif tokens[i].token_type == TokenType.SEE:
            if (
                i + 3 < len(tokens) and
                tokens[i + 1].token_type == TokenType.ALSO and
                tokens[i + 2].token_type == TokenType.NEWLINE and
                tokens[i + 3].token_type == TokenType.HEADER
            ):
                return 4
            else:
                return 0
        elif tokens[i].token_type in KEYWORDS:
            if (
                i + 2 < len(tokens) and
                tokens[i + 1].token_type == TokenType.NEWLINE and
                tokens[i + 2].token_type == TokenType.HEADER
            ):
                return 3
            else:
                return 0
        else:
            return 0

    # Handle an empty docstring.
    if not tokens:
        return list()

    curr = list()  # type: List[Token]
    overall = list()  # type: List[List[Token]]
    i = 0
    while i < len(tokens) and tokens[i].token_type != TokenType.NEWLINE:
        i += 1
    while i < len(tokens) and tokens[i].token_type == TokenType.NEWLINE:
        i += 1
    overall = [tokens[:i]]
    while i < len(tokens):
        beginning_tokens = at_section_boundary(i)
        if beginning_tokens and curr:
            overall.append(curr)
            curr = list()
            curr.extend(tokens[i:i + beginning_tokens + 1])
            i += beginning_tokens + 1
        else:
            curr.append(tokens[i])
            i += 1
    if curr:
        overall.append(curr)
    return overall


def _match(token):
    # type: (Token) -> List[Union[Callable, BaseGrammar]]
    """Match the given token from the given section to a set of grammars.

    Args:
        token: The token to match.  This should hint at what sections
            could possibly be here.

    Returns:
        A list of grammars to be tried in order.

    """
    tt_lookup = {
        TokenType.RETURNS: [
            ReturnsGrammar,
            long_description_parse,
        ],
        TokenType.ARGUMENTS: [
            ArgumentsGrammar,
            long_description_parse,
        ],
        TokenType.YIELDS: [
            YieldsGrammar,
            long_description_parse,
        ],
        TokenType.RAISES: [
            RaisesGrammar,
            long_description_parse,
        ],
        TokenType.WARNS: [
            WarnsGrammar,
            long_description_parse,
        ],
        TokenType.RECEIVES: [
            ReceivesGrammar,
            long_description_parse,
        ],
        TokenType.OTHER: [
            OtherArgumentsGrammar,
            long_description_parse,
        ],
        # Discard these two sections -- there's nothing
        # to check against the function description.
        TokenType.SEE: [
            long_description_parse,
        ],
        TokenType.NOTES: [
            long_description_parse,
        ],
        TokenType.EXAMPLES: [
            long_description_parse,
        ],
    }  # type: Dict[TokenType, List[Union[BaseGrammar, Callable]]]  # noqa: E501
    return tt_lookup.get(token.token_type, [long_description_parse])


def lookup(section, section_index=-1):
    Assert(len(section) > 0, 'Expected a non-empty section.')
    grammars = _match(section[0])
    if section_index == 0:
        return [ShortDescriptionGrammar] + grammars
    return grammars


def combinator(*args):
    def inner(*nodes):
        if len(nodes) == 1:
            return CykNode(
                symbol='docstring',
                lchild=nodes[0],
            )
        elif len(nodes) == 2:
            return CykNode(
                symbol='docstring',
                lchild=nodes[0],
                rchild=nodes[1],
            )
    if args:
        return reduce(inner, args)
    else:
        # The arguments are empty, so we return an
        # empty docstring.
        return CykNode(symbol='docstring')


def parse(tokens):
    # type: (List[Token]) -> Optional[CykNode]
    def mapped_lookup(section, section_index=-1):
        for grammar in lookup(section, section_index):
            if inspect.isclass(grammar):
                yield lambda x: cyk_parse(grammar, x)
            else:
                yield grammar
    return parser_combinator(top_parse, mapped_lookup, combinator, tokens)



"""Defines a base class far describing grammars."""

import abc
from typing import (  # noqa: F401
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from ..custom_assert import Assert
from ..token import TokenType

Annotation = Any # TODO(000): This should actually be Union[Identifier, DarglintError]
NonTerminalDerivation = Tuple[List[Annotation], str, str, int]
TerminalDerivation = Tuple[TokenType, int]
Derivation = Union[NonTerminalDerivation, TerminalDerivation]


class Production(object):
    """Represents a production in a grammar."""

    def __init__(self, lhs, *rhs, annotations=list()):
        # type: (str, Derivation, Optional[List[Any]]) -> None
        """Create a new production.

        Args:
            lhs: The left-hand side of the production.
            *rhs: The items in the right-hand side of the
                production.
            annotations: A list of annotations.  Here, they should
                be the class of errors which are to be raised as a
                result of the given production being parsed.

        """
        self.lhs = lhs
        self.rhs = rhs
        self.annotations = annotations

    @classmethod
    def with_annotations(cls, lhs, annotations, *rhs):
        return cls(lhs, annotations=annotations, *rhs)


# Convenience name.
P = Production


class BaseGrammar(abc.ABC):

    @property
    @abc.abstractmethod
    def productions(self):
        # type: () -> List[Production]
        pass

    @property
    @abc.abstractmethod
    def start(self):
        # type: () -> str
        pass

    @classmethod
    def get_symbol_lookup(cls):
        # type: () -> Dict[str, int]
        lookup = dict()  # type: Dict[str, int]

        # We have to tell the type checker that productions is, in fact,
        # a list, since enumerate actually takes a wider type (Iterable),
        # and it gets confused.
        Assert(
            isinstance(cls.productions, list),
            'Expected productions to be a list.'
        )
        if not isinstance(cls.productions, list):
            return dict()
        for i, production in enumerate(cls.productions):
            symbol = production.lhs
            lookup[symbol] = i
        return lookup

    @classmethod
    def to_dot(cls):
        # () -> str

        def normalize(name):
            return name.replace('-', '_').replace('.', '_')

        def interpolate(curr, max_weight):
            min_color = (200, 200, 200)
            max_color = (0, 0, 0)
            return tuple(
                int(u - (u - l) * (curr / max_weight))
                for l, u in zip(max_color, min_color)
            )

        def to_hex(color):
            return '#{}{}{}'.format(*[
                format(x, '02x') for x in color
            ])

        start = cls.start
        if not start:
            return ''

        max_weight = 0
        for production in cls.productions:
            for derivation in production.rhs:
                if len(derivation) == 4:
                    if derivation[-1] > max_weight:
                        max_weight = derivation[-1]

        join_count = 0
        ret = 'digraph G {\n'
        for production in cls.productions:
            lhs = normalize(production.lhs)
            for derivation in production.rhs:
                if len(derivation) == 2:
                    rhs = normalize(str(derivation[0]))
                    color = to_hex(interpolate(0, max_weight))
                    ret += (
                        '  {} -> {} [color="{}"];\n\n'.format(
                            lhs,
                            rhs,
                            color,
                        )
                    )
                elif len(derivation) == 4:
                    weight = derivation[-1]
                    color = to_hex(interpolate(weight, max_weight))
                    join_node = '_join{}'.format(join_count)
                    join_count += 1
                    annotations = [x.__name__ for x in derivation[0]]
                    if annotations:
                        node_color = '#c8e6c9'
                        if any(['error' in x.lower() for x in annotations]):
                            node_color = '#ffcdd2'
                        ret += (
                            '  {} [label="{}", shape="rect", '
                            'style="filled", fillcolor="{}"]\n'
                        ).format(
                            join_node,
                            '\\n'.join(annotations),
                            node_color,
                        )
                    else:
                        ret += (
                            '  {} [label="", color="none", '
                            'height="0", width="0"];\n'
                        ).format(join_node)
                    ret += (
                        '  {} -> {} [arrowhead="none", color="{}"];\n'
                    ).format(
                        lhs,
                        join_node,
                        color,
                    )
                    ret += (
                        '  {} -> {{ {}, {} }} [color="{}"];\n\n'
                    ).format(
                        join_node,
                        normalize(derivation[1]),
                        normalize(derivation[2]),
                        color,
                    )
        ret += '}'
        return ret



import enum


class Sections(enum.Enum):
    SHORT_DESCRIPTION = 0
    LONG_DESCRIPTION = 1
    ARGUMENTS_SECTION = 2
    RAISES_SECTION = 4
    YIELDS_SECTION = 6
    RETURNS_SECTION = 8
    VARIABLES_SECTION = 10
    NOQAS = 13




import ast
from typing import (
    Any,
    Dict,
    List,
)


class ArgumentVisitor(ast.NodeVisitor):
    """Reports which arguments a function contains."""

    def __init__(self, *args, **kwargs):
        # type: (List[Any], Dict[str, Any]) -> None

        # https://github.com/python/mypy/issues/5887
        super(ArgumentVisitor, self).__init__(*args, **kwargs)  # type: ignore

        # The arguments found in the function.
        self.arguments = list()  # type: List[str]
        self.types = list()  # type: List[str]

    def add_arg_by_name(self, name, arg):
        self.arguments.append(name)
        if arg.annotation is not None and hasattr(arg.annotation, 'id'):
            self.types.append(arg.annotation.id)
        else:
            self.types.append(None)

    def visit_arguments(self, node):
        # type: (ast.arguments) -> ast.AST
        if hasattr(node, 'posonlyargs'):
            for arg in node.posonlyargs:
                self.add_arg_by_name(arg.arg, arg)

        for arg in node.args:
            self.add_arg_by_name(arg.arg, arg)

        for arg in node.kwonlyargs:
            self.add_arg_by_name(arg.arg, arg)

        # Handle single-star arguments.
        if node.vararg is not None:
            name = '*' + node.vararg.arg
            self.add_arg_by_name(name, node.vararg)

        if node.kwarg is not None:
            name = '**' + node.kwarg.arg
            self.add_arg_by_name(name, node.kwarg)
        return self.generic_visit(node)



from abc import ABC, abstractmethod
from typing import (
    ClassVar,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    Iterable,
)

from .sections import Sections
from ..strictness import Strictness


class BaseDocstring(ABC):
    """The interface for a docstring object which can be used with checkers.

    Unfortunately, slight differences in the parsers result
    in kind of inconsistent values for different nodes. (Sometimes
    there are blank lines and such.)  For that reason, we include some
    tests to try to make the output from the below three methods as
    consistent as possible.

    """

    supported_sections = tuple(Sections) # type: ClassVar[Tuple[Sections, ...]]

    @abstractmethod
    def get_section(self, section):
        # type: (Sections) -> Optional[str]
        """Get an entire section of the docstring.

        Args:
            section: The section to get.

        Raises:
            Exception: If the section type is unsupported.

        Returns:
            A string representing the section as a contiguous piece,
            or None if the section was not there.

        # noqa: I202
        # noqa: I402

        """
        pass

    @abstractmethod
    def get_types(self, section):
        # type: (Sections) -> Optional[Union[str, List[Optional[str]]]]
        """Get the type of the section, or of the items in the section.

        Args:
            section: The section whose type or types we are getting.
                If the section is a returns section, for example, we get
                the return type (or none if not specified).  If it is an
                arguments section, we get the types for each item.

        Raises:
            Exception: If the given section does not have a type.
                (For example, the short description.)

        Returns:
            The type or types of the section, or None if it's not
            defined.

        # noqa: I202
        # noqa: I402

        """
        pass

    @abstractmethod
    def get_items(self, section):
        # type: (Sections) -> Optional[List[str]]
        """Get the item names in the section.

        Args:
            section: The section whose item names we are getting.

        Raises:
            Exception: If the given section is not composed of
                items (e.g. a returns section, or a long description.)

        Returns:
            The items in the section (or None of the section is
            not defined.)

        # noqa: I202
        # noqa: I402

        """
        pass

    @abstractmethod
    def get_style_errors(self):
        # type: () -> Iterable[Tuple[Callable, Tuple[int, int]]]
        """Get any style errors annotated on the tree.

        Yields:
            Instances of DarglintErrors for style issues.

        # noqa: I302

        """
        pass

    @abstractmethod
    def get_noqas(self):
        # type: () -> Dict[str, List[str]]
        pass

    @abstractmethod
    def get_line_numbers(self, node_type):
        # type: (str) -> Optional[Tuple[int, int]]
        pass

    @abstractmethod
    def get_line_numbers_for_value(self, node_type, value):
        # type: (str, str) -> Optional[Tuple[int, int]]
        pass

    @property
    @abstractmethod
    def ignore_all(self):
        # type: () -> bool
        pass

    def satisfies_strictness(self, strictness):
        # type(Strictness) -> bool
        """Return true if the docstring has no more than the min strictness.

        Args:
            strictness: The minimum amount of strictness which should
                be present in the docstring.

        Returns:
            True if there is no more than the minimum amount of strictness.

        """
        sections = {
            section
            for section in self.supported_sections
            if self.get_section(section)
        }
        if strictness == Strictness.SHORT_DESCRIPTION:
            return sections == {Sections.SHORT_DESCRIPTION}
        elif strictness == Strictness.LONG_DESCRIPTION:
            return sections in (
                {Sections.SHORT_DESCRIPTION},
                # Shouldn't be possible, but if it is in the future, then
                # we should allow this.
                {Sections.LONG_DESCRIPTION},
                {Sections.SHORT_DESCRIPTION, Sections.LONG_DESCRIPTION},
            )
        else:
            return False



from .raise_visitor import RaiseVisitor
from .return_visitor import ReturnVisitor
from .yield_visitor import YieldVisitor
from .function_scoped_visitor import FunctionScopedVisitorMixin
from .argument_visitor import ArgumentVisitor
from .assert_visitor import AssertVisitor
from .variable_visitor import VariableVisitor
from .abstract_callable_visitor import AbstractCallableVisitor


# ATTENTION: FunctionScopedVisitorMixin needs to be first,
# otherwise it is not able to stop descending into wrapped
# functions.
class AnalysisVisitor(FunctionScopedVisitorMixin,
                      AbstractCallableVisitor,
                      RaiseVisitor,
                      YieldVisitor,
                      ArgumentVisitor,
                      VariableVisitor,
                      ReturnVisitor,
                      AssertVisitor):
    """Finds attributes which should be part of the function signature."""
    pass



"""Defines a custom assert function for darglint."""

from .config import (
    get_logger,
    get_config,
    AssertStyle,
)

from typing import (
    Any,
    Optional,
)


def Assert(expr, message):  # type: (Any, Optional[str]) -> None
    """Asserts that the given expression is true.

    Args:
        expr: The expression to evaluate.  It will be
            interpreted as a boolean.
        message: A message describing the expectation of
            this assertion, describing the error encountered,
            or some other debugging information.

    Raises:
        AssertionError: If darglint is configured to raise
            on failed assertions, otherwise logs.

    """
    if expr:
        return

    style = get_config().assert_style
    if style == AssertStyle.RAISE:
        raise AssertionError(message)
    elif style == AssertStyle.LOG:
        logger = get_logger()
        logger.error(message or 'Failed assertion')



import ast
from typing import (
    Iterable,
    Union
)

def _has_decorator(function, decorators):
    # # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], Union[str, Iterable[str]]) -> bool
    if isinstance(decorators, str):
        decorators = (decorators,)

    for decorator in function.decorator_list:
        # Attributes (setters and getters) won't have an id.
        if isinstance(decorator, ast.Name) and decorator.id in decorators:
            return True
    return False



from enum import Enum


class Strictness(Enum):
    """The minimum strictness with which to apply checks.

    Strictness does not describe whether or not a check
    should be applied. Rather, if a check is done, strictness
    describes how intense/strict/deep the check should be.

    Each level here describes what is required of the
    docstring at the given level of strictness.  For example,
    SHORT_DESCRIPTION describes the situation where one-liners are
    allowed, and sections are not required.

    If the docstring being checked contains more than the
    allowed amount below, then it is assumed that everything
    must be checked.

    """

    # Allow a single-line description.
    SHORT_DESCRIPTION = 1

    # Allow a single-line description followed by a long
    # description, but no sections.
    LONG_DESCRIPTION = 2

    # Require everything.
    FULL_DESCRIPTION = 3

    @classmethod
    def from_string(cls, strictness):
        strictness = strictness.lower().strip()
        if strictness in {'short_description', 'short'}:
            return cls.SHORT_DESCRIPTION
        if strictness in {'long_description', 'long'}:
            return cls.LONG_DESCRIPTION
        if strictness in {'full_description', 'full'}:
            return cls.FULL_DESCRIPTION

        raise Exception(
            'Unrecognized strictness amount "{}".  '.format(strictness) +
            'Should be one of {"short", "long", "full"}'
        )



"""The entry point for flake8."""

import ast  # noqa
from typing import (  # noqa
    Iterator,
    Tuple,
)

from .docstring.style import DocstringStyle
from .function_description import (
    get_function_descriptions,
)
from .integrity_checker import IntegrityChecker
from .config import (
    Configuration,
    get_config,
)
from .strictness import Strictness


__version__ = '1.8.0'


class DarglintChecker(object):

    name = 'flake8-darglint'
    version = __version__
    config = get_config()

    def __init__(self, tree, filename):
        self.tree = tree
        self.filename = filename
        self.verbosity = 2

    def run(self):
        # type: () -> Iterator[Tuple[int, int, str, type]]
        if '*' in self.config.ignore:
            return

        # Remember the last line number, so that if there is an
        # exception raised by Darglint, we can at least give a decent
        # idea of where it was raised.
        last_line = 1
        try:
            functions = get_function_descriptions(self.tree)
            checker = IntegrityChecker(
                raise_errors=False,
            )
            checker.config = self.config
            for function in functions:
                checker.run_checks(function)

            error_report = checker.get_error_report(
                self.verbosity,
                self.filename
            )
            for line, col, msg in error_report.flake8_report():
                last_line = line
                yield (line, col, msg, type(self))

        except Exception as ex:
            yield (
                last_line,
                0,
                'DAR000: Unexpected exception in darglint: ' + str(ex),
                type(self)
            )

    @classmethod
    def add_options(cls, option_manager):
        defaults = cls.config

        option_manager.add_option(
            '--docstring-style',
            default=defaults.style.name,
            parse_from_config=True,
            help='Docstring style to use for Darglint',
        )

        option_manager.add_option(
            '--strictness',
            default=defaults.strictness.name,
            parse_from_config=True,
            help='Strictness level to use for Darglint',
        )

    @classmethod
    def parse_options(cls, options):
        cls.config.style = DocstringStyle.from_string(options.docstring_style)
        cls.config.strictness = Strictness.from_string(options.strictness)



import ast

from typing import (
    Any,
)

class FunctionScopedVisitorMixin(ast.NodeVisitor):
    """A visitor which is scoped to a single function.

    This visitor assumes that its `visit` method is called
    on a `ast.FunctionDef`, and it will not examine nested
    functions.

    """

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None

        # TODO: https://github.com/python/mypy/issues/4001
        super(FunctionScopedVisitorMixin, self).__init__(*args, **kwargs)  # type: ignore  # noqa: E501

        # Whether we have passed the initial `FunctionDef` node.
        self.in_function = False

    def visit_Lambda(self, node):
        # type: (ast.Lambda) -> ast.AST
        if not self.in_function:
            self.in_function = True
            return getattr(
                super(),
                "visit_Lambda",
                super().generic_visit
            )(node)
        else:
            # Return a synthetic Pass node, to make type checking happy
            # (and to not violate the contract.)  Since it has no children,
            # it will effectively stop the visit.
            return ast.Pass()

    def visit_FunctionDef(self, node):
        # type: (ast.FunctionDef) -> ast.AST
        if not self.in_function:
            self.in_function = True
            return getattr(
                super(),
                "visit_FunctionDef",
                super().generic_visit
            )(node)
        else:
            return ast.Pass()

    def visit_AsyncFunctionDef(self, node):
        # type: (ast.AsyncFunctionDef) -> ast.AST
        if not self.in_function:
            self.in_function = True
            return getattr(
                super(),
                "visit_AsyncFunctionDef",
                super().generic_visit
            )(node)
        else:
            return ast.Pass()



import ast
from typing import (
    Any,
    Dict,
    List,
)


class VariableVisitor(ast.NodeVisitor):

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None

        # TODO: https://github.com/python/mypy/issues/4001
        super(VariableVisitor, self).__init__(*args, **kwargs)  # type: ignore
        self.variables = list()  # type: List[ast.Name]

    def visit_Name(self, node):
        # type: (ast.Name) -> ast.AST
        # Only gather names during assignment.  Others are unnecessary,
        # and could be from a different context.
        if hasattr(node, 'ctx') and isinstance(node.ctx, ast.Store):
            self.variables.append(node)
        return self.generic_visit(node)



import ast
from typing import (
    Any,
    Dict,
    List,
)

from .analysis_helpers import (
    _has_decorator
)

class AbstractCallableVisitor(ast.NodeVisitor):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_abstract = None

    def _is_docstring(self, node):
        # type: (ast.AST) -> bool
        return (
            isinstance(node, ast.Expr) and (
                (
                    isinstance(node.value, ast.Constant) and
                    isinstance(node.value.value, str)
                ) or (
                    isinstance(node.value, ast.Str) # Python < 3.8
                )
            )
        )

    def _is_ellipsis(self, node):
        # type: (ast.AST) -> bool

        return (
            isinstance(node, ast.Expr) and (
                (
                    isinstance(node.value, ast.Constant) and
                    node.value.value is Ellipsis
                ) or (
                    isinstance(node.value, ast.Ellipsis) # Python < 3.8
                )
            )
        )

    def _is_raise_NotImplementedException(self, node):
        # type: (ast.AST) -> bool
        return (
            isinstance(node, ast.Raise) and (
                (
                    isinstance(node.exc, ast.Name) and
                    node.exc.id == "NotImplementedError"
                ) or (
                    isinstance(node.exc, ast.Call)
                    and isinstance(node.exc.func, ast.Name)
                    and node.exc.func.id == "NotImplementedError"
                )
            )
        )

    def _is_return_NotImplemented(self, node):
        # type: (ast.AST) -> bool
        return (
            isinstance(node, ast.Return) and
            isinstance(node.value, ast.Name) and
            node.value.id == "NotImplemented"
        )

    def analyze_pure_abstract(self, node):
        # type: (ast.AST) -> bool
        assert isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)), (
            "Assuming this analysis is only called on functions"
        )

        if not _has_decorator(node, "abstractmethod"):
            return False

        children = len(node.body)

        # maximum docstring and one statement
        if children > 2:
            return False

        if children == 2:
            if not self._is_docstring(node.body[0]):
                return False

            statement = node.body[1]
        else:
            statement = node.body[0]

        if (
            isinstance(statement, ast.Pass) or
            self._is_ellipsis(statement) or
            self._is_raise_NotImplementedException(statement) or
            self._is_return_NotImplemented(statement) or
            (
                children == 1 and
                self._is_docstring(statement)
            )
        ):
            return True

        return False

    def visit_FunctionDef(self, node):
        # type: (ast.FunctionDef) -> ast.AST
        self.is_abstract = self.analyze_pure_abstract(node)
        return self.generic_visit(node)

    def visit_AsyncFunctionDef(self, node):
        # type: (ast.AsyncFunctionDef) -> ast.AST
        self.is_abstract = self.analyze_pure_abstract(node)
        return self.generic_visit(node)



from collections import (
    deque,
)
from typing import (
    Any,
    Iterator,
    Optional,
    List,
    Tuple,
)

from .token import (
    Token,
    TokenType,
)

WHITESPACE = {TokenType.INDENT, TokenType.NEWLINE}


# A best guess at the maximum height of a docstring tree,
# for use in recursion bounds checking.
MAX_TREE_HEIGHT = 300


class CykNode(object):
    """A node for use in a cyk parse."""

    def __init__(self,
                 symbol,
                 lchild=None,
                 rchild=None,
                 value=None,
                 annotations=list(),
                 weight=0):
        # type: (str, Optional[CykNode], Optional[CykNode], Optional[Token], List[Any], int) -> None  # noqa: E501
        self.symbol = symbol
        self.lchild = lchild
        self.rchild = rchild
        self.value = value
        self.annotations = annotations
        self._line_number_cache = None  # type: Optional[Tuple[int, int]]

        # If there is an explicit weight, we definitely want to use
        # that (there was probably a good reason it was given.)
        #
        # If no weight was given, but the children have weights, then
        # we probably want to give preference to this node over a node
        # which has no weights at all.
        if weight:
            self.weight = weight
        else:
            self.weight = max([
                0,
                self.lchild.weight if self.lchild else 0,
                self.rchild.weight if self.rchild else 0,
            ])

    def __repr__(self):
        if hasattr(self.value, 'token_type'):
            return '<{}: {}>'.format(
                self.symbol,
                str(self.value.token_type)[10:] if self.value else '',
            )
        return '<{}>'.format(self.value)

    def __str__(self, indent=0):
        if self.value:
            ret = (
                ' ' * indent
                + str(self.value.token_type)
                + ': '
                + repr(self.value.value)
            )
        else:
            ret = ' ' * indent + self.symbol
        if self.annotations:
            ret += ': ' + ', '.join([str(x) for x in self.annotations])
        if self.lchild:
            ret += '\n' + self.lchild.__str__(indent + 2)
        if self.rchild:
            ret += '\n' + self.rchild.__str__(indent + 2)
        return ret

    def in_order_traverse(self):
        # type: () -> Iterator[CykNode]
        if self.lchild:
            yield from self.lchild.in_order_traverse()
        yield self
        if self.rchild:
            yield from self.rchild.in_order_traverse()

    def breadth_first_walk(self):
        queue = deque([self])
        while queue:
            curr = queue.pop()
            yield curr
            if curr.lchild:
                queue.appendleft(curr.lchild)
            if curr.rchild:
                queue.appendleft(curr.rchild)

    def first_instance(self, symbol):
        # type: (str) -> Optional['CykNode']
        for node in self.breadth_first_walk():
            if node.symbol == symbol:
                return node
        return None

    def walk(self):
        # type: () -> Iterator['CykNode']
        yield from self.in_order_traverse()

    def equals(self, other):
        # type: (Optional['CykNode']) -> bool
        if other is None:
            return False
        if self.symbol != other.symbol:
            return False
        if self.value != other.value:
            return False
        if self.lchild and not self.lchild.equals(other.lchild):
            return False
        if self.rchild and not self.rchild.equals(other.rchild):
            return False
        return True

    def reconstruct_string(self, strictness=0):
        # type: (int) -> str
        """Reconstruct the docstring.

        This method should rebuild the docstring while fixing style
        errors.  The errors themselves determine how to fix the node
        they apply to.  (If there isn't a good fix, then it's just the
        identity function.)

        Args:
            strictness: How strictly we should correct.  If an error
                doesn't match the strictness, we won't correct for
                that error.

        Returns:
            The docstring, reconstructed.

        """
        # In order to make a reasonable guess as to the whitespace
        # to apply between characters, we use a 3-token sliding
        # window.
        window_size = 3
        window = deque(maxlen=window_size)  # type: deque
        source = self.in_order_traverse()

        # Fill the buffer.
        while len(window) < window_size:
            try:
                node = next(source)
            except StopIteration:
                break
            if node.value:
                window.append(node.value)

        if not window:
            return ''

        ret = window[0].value

        # Slide the window, filling the return value.
        while len(window) > 1:
            is_whitespace = (
                window[0].token_type in WHITESPACE
                or window[1].token_type in WHITESPACE
            )
            is_colon = window[1].token_type == TokenType.COLON
            if is_whitespace or is_colon:
                ret += window[1].value
            else:
                ret += ' ' + window[1].value

            found_token = False
            for node in source:
                if node.value:
                    window.append(node.value)
                    found_token = True
                    break
            if not found_token:
                break

        if len(window) == 3:
            if (window[1].token_type in WHITESPACE
                    or window[2].token_type in WHITESPACE
                    or window[2].token_type == TokenType.COLON):
                ret += window[2].value
            else:
                ret += ' ' + window[2].value

        return ret

    def _get_line_numbers_cached(self, recurse=0):
        # type: (int) -> Tuple[int, int]
        if recurse > MAX_TREE_HEIGHT:
            return (-1, -1)
        if self.value:
            return (self.value.line_number, self.value.line_number)
        elif self._line_number_cache:
            return self._line_number_cache
        leftmost = -1
        if self.lchild:
            leftmost = self.lchild._get_line_numbers_cached(recurse + 1)[0]
        rightmost = leftmost
        if self.rchild:
            rightmost = self.rchild._get_line_numbers_cached(recurse + 1)[1]
        self._line_number_cache = (leftmost, rightmost)
        return self._line_number_cache or (-1, -1)

    @property
    def line_numbers(self):
        # type: () -> Tuple[int, int]
        return self._get_line_numbers_cached()



"""Utilities for debugging Darglint.

This file should not be imported into Darglint,
and should ideally be excluded from the sources.

"""
from ast import (
    AST,
)
from collections import (
    deque,
)
from typing import (
    Any,
    List,
    Optional,
    Set,
)
from .node import CykNode
from .config import (
    get_config,
    Configuration,
)


class AstNodeUtils(object):

    @staticmethod
    def iter_fields(node):
        for field in node._fields:
            try:
                yield field, getattr(node, field)
            except AttributeError:
                pass

    @staticmethod
    def _get_node_label(node):
        if hasattr(node, 'id'):
            return r'{}\nid: {}'.format(
                node.__class__.__name__,
                node.id,
            )
        elif hasattr(node, 'attr') and isinstance(node.attr, str):
            return r'{}\nattr: {}'.format(
                node.__class__.__name__,
                node.attr,
             )
        else:
            return node.__class__.__name__

    @staticmethod
    def to_dot(node):
        # type: (AST) -> str
        dot = ['digraph G {']  # type: List[str]
        id = 0
        queue = deque([(id, node)])
        while queue:
            parent_id, curr = queue.pop()
            dot.insert(1, '  {}_{} [label="{}"];'.format(
                curr.__class__.__name__,
                parent_id,
                AstNodeUtils._get_node_label(curr),
            ))
            for field, value in AstNodeUtils.iter_fields(curr):
                if isinstance(value, list):
                    for item in value:
                        if isinstance(item, AST):
                            id += 1
                            queue.appendleft((id, item))
                            dot.append('  {}_{} -> {}_{};'.format(
                                curr.__class__.__name__, parent_id,
                                item.__class__.__name__, id,
                            ))
                elif isinstance(value, AST):
                    id += 1
                    queue.appendleft((id, value))
                    dot.append('  {}_{} -> {}_{};'.format(
                        curr.__class__.__name__, parent_id,
                        value.__class__.__name__, id,
                    ))
        dot.append('}')
        return '\n'.join(dot)


class CykNodeUtils(object):

    @staticmethod
    def contains(self, symbol=None, value=None):
        # type: (CykNode, Optional[str], Optional[str]) -> bool
        """Return true if the tree contains the given symbol.

        This is intended only for testing.

        Args:
            self: The node where we should begin the search.
            symbol: The symbol to search for.
            value: If defined, the string value the node should
                have in order to match.

        Returns:
            True if the symbol/value is in the tree, false otherwise.

        """
        def _match(n):
            symbol_match = symbol is None or n.symbol == symbol
            value_match = value is None or n.value == value
            return symbol_match and value_match
        for node in self.walk():
            if _match(node):
                return True
        return False

    @staticmethod
    def get_annotated(self, annotation):
        # type: (CykNode, Any) -> Set[CykNode]
        nodes = set()
        for node in self.walk():
            if annotation in node.annotations:
                nodes.add(node)
        return nodes

    @staticmethod
    def to_dot(cyk_node, is_root=True, encountered=set()):
        # type: (CykNode, bool, Set[str]) -> str
        def _get_name(node):
            i = 0
            name = node.symbol
            if node.value:
                if node.value.value:
                    name += '_' + node.value.value
                else:
                    name += '_' + str(node.value)
            for x in '.*!@#$%^*,- ;:\'"\n\t{}[]()':
                name = name.replace(x, '_')
            while name.replace('__', '_') != name:
                name = name.replace('__', '_')
            name = name.lower()
            while name + str(i) in encountered:
                i += 1
            return name + str(i)

        def _get_value(node):
            if cyk_node.value and cyk_node.value.value:
                if cyk_node.value.value.isspace():
                    return repr(cyk_node.value.value).replace('\\', '\\\\')
                return cyk_node.value.value
            elif cyk_node.value:
                return str(cyk_node.value)
            else:
                return cyk_node.symbol

        if is_root:
            ret = 'digraph G {\n'
        else:
            ret = ''

        # Print this node's relationship with its children.
        name = _get_name(cyk_node)
        encountered.add(name)
        if cyk_node.lchild or cyk_node.rchild:
            ret += name + (
                '  [shape="oval", style="filled", '
                ' label="{}",'
                ' fillcolor="#fffde7"];\n'
            ).format(_get_value(cyk_node))
            if cyk_node.lchild:
                childname = _get_name(cyk_node.lchild)
                ret += '{} -> {}'.format(name, childname)
                if cyk_node.annotations:
                    ret += '[label="'
                    for annotation in cyk_node.annotations:
                        ret += annotation.__name__ + ',\\n'
                    ret += '"]'
                ret += ';\n'
            if cyk_node.rchild:
                childname = _get_name(cyk_node.rchild)
                if cyk_node.lchild and _get_name(cyk_node.lchild) == childname:
                    childname += '_'
                ret += '{} -> {};\n'.format(name, childname)
        else:
            ret += name + (
                ' [shape="rectangle", style="filled",'
                ' label="{}"'
                ' fillcolor="#80deea"];\n'
            ).format(_get_value(cyk_node))

        # Add all of the children's relationships.
        if cyk_node.lchild:
            ret += CykNodeUtils.to_dot(
                cyk_node.lchild, False, encountered
            ) + '\n'
        if cyk_node.rchild:
            ret += CykNodeUtils.to_dot(
                cyk_node.rchild, False, encountered
            ) + '\n'

        if is_root:
            ret += '}'
        return ret


class ConfigurationContext(object):
    """A context manager for the configuration.

    Resets the configuration to the value it had prior
    to entering the context.

    """

    def __init__(self, **kwargs):
        self.original = dict()
        self.config = get_config()
        self.kwargs = kwargs

    def __enter__(self):
        # Save the state of the original item.
        for keyword, value in vars(self.config).items():
            self.original[keyword] = getattr(self.config, keyword)

        # Override with the default configuration.
        default_config = Configuration.get_default_instance()
        for keyword in self.original:
            setattr(self.config, keyword, getattr(default_config, keyword))

        # Apply test-specific values.
        for keyword, value in self.kwargs.items():
            setattr(self.config, keyword, value)

        return self.config

    def __exit__(self, exc_type, exc_value, exc_traceack):
        for keyword, value in self.original.items():
            setattr(self.config, keyword, value)



import ast
from typing import (
    List,
    Set,
    Union,
)

from .analysis_helpers import (
    _has_decorator
)


class FunctionAndMethodVisitor(ast.NodeVisitor):

    def __init__(self):
        # type: () -> None
        self.callables = set()  # type: Set[Union[ast.FunctionDef, ast.AsyncFunctionDef]]
        self._methods = set()  # type: Set[Union[ast.FunctionDef, ast.AsyncFunctionDef]]
        self._properties = set()  # type: Set[Union[ast.FunctionDef, ast.AsyncFunctionDef]]

    @property
    def functions(self):
        # type: () -> List[Union[ast.FunctionDef, ast.AsyncFunctionDef]]
        return list(self.callables - self._methods - self._properties)

    @property
    def methods(self):
        # type: () -> List[Union[ast.FunctionDef, ast.AsyncFunctionDef]]
        return list(self._methods)

    @property
    def properties(self):
        # type: () -> List[Union[ast.FunctionDef, ast.AsyncFunctionDef]]
        return list(self._properties)

    def visit_ClassDef(self, node):
        # type: (ast.ClassDef) -> ast.AST
        for item in node.body:
            if isinstance(item, ast.FunctionDef) or isinstance(
                item, ast.AsyncFunctionDef
            ):
                if _has_decorator(item, "property"):
                    self._properties.add(item)
                else:
                    self._methods.add(item)
        return self.generic_visit(node)

    def visit_FunctionDef(self, node):
        # type: (ast.FunctionDef) -> ast.AST
        self.callables.add(node)
        return self.generic_visit(node)

    def visit_AsyncFunctionDef(self, node):
        # type: (ast.AsyncFunctionDef) -> ast.AST
        self.callables.add(node)
        return self.generic_visit(node)



"""A linter for docstrings following the google docstring format."""
import ast
from collections import deque
import sys
from enum import Enum
from typing import (
    Callable,
    Iterator,
    List,
    Set,
    Tuple,
    Optional,
    Union,
    Type,
    Any,
)

from .analysis.analysis_visitor import (
    AnalysisVisitor,
)
from .analysis.function_and_method_visitor import (
    FunctionAndMethodVisitor,
)
from .config import get_logger
from .analysis.analysis_helpers import (
    _has_decorator
)


logger = get_logger()


FunctionDef = ast.FunctionDef  # type: Union[Type[Any], Tuple[Type[Any], Type[Any]]]  # noqa: E501
if hasattr(ast, 'AsyncFunctionDef'):
    FunctionDef = (ast.FunctionDef, ast.AsyncFunctionDef)


def read_program(filename):  # type: (str) -> Union[bytes, str]
    """Read a program from a file.

    Args:
        filename: The name of the file to read. If set to '-', then we will
            read from stdin.

    Returns:
        The program as a single string.

    """
    program = None  # type: Union[bytes, Optional[str]]
    if filename == '-':
        program = sys.stdin.read()
    else:
        with open(filename, 'rb') as fin:
            program = fin.read()
    return program or ''


def _get_docstring(fun):  # type: (ast.AST) -> Optional[str]
    return ast.get_docstring(fun)


def _get_all_functions(tree):  # type: (ast.AST) -> Iterator[Union[ast.FunctionDef, ast.AsyncFunctionDef]]  # noqa: E501
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            yield node
        elif hasattr(ast, 'AsyncFunctionDef'):
            if isinstance(node, ast.AsyncFunctionDef):
                yield node


def _get_all_classes(tree):  # type: (ast.AST) -> Iterator[ast.ClassDef]
    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef):
            yield node


def _get_all_methods(tree):  # type: (ast.AST) -> Iterator[Union[ast.FunctionDef, ast.AsyncFunctionDef]]  # noqa: E501
    for klass in _get_all_classes(tree):
        for fun in _get_all_functions(klass):
            yield fun


def _get_return_type(fn):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Optional[str]
    if fn.returns is not None and hasattr(fn.returns, 'id'):
        return getattr(fn.returns, 'id')
    return None


def get_line_number_from_function(fn):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> int
    """Get the line number for the end of the function signature.

    The function signature can be farther down when the parameter
    list is split across multiple lines.

    Args:
        fn: The function from which we are getting the line number.

    Returns:
        The line number for the start of the docstring for this
        function.

    """
    line_number = fn.lineno
    if hasattr(fn, 'args') and fn.args.args:
        last_arg = fn.args.args[-1]
        line_number = last_arg.lineno
    return line_number


class FunctionType(Enum):

    FUNCTION = 1
    METHOD = 2
    PROPERTY = 3


class FunctionDescription(object):
    """Describes a function or method.

    Whereas a `Docstring` object describes a function's docstring,
    a `FunctionDescription` describes the function itself.  (What,
    ideally, the docstring should describe.)

    """

    def __init__(self, function_type, function):
        # type: (FunctionType, Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> None
        """Create a new FunctionDescription.

        Args:
            function_type: Type of the function.
            function: The base node of the function.

        """
        self.is_method = (function_type == FunctionType.METHOD)
        self.is_property = (function_type == FunctionType.PROPERTY)
        self.function = function
        self.line_number = get_line_number_from_function(function)
        self.name = function.name
        visitor = AnalysisVisitor()
        try:
            visitor.visit(function)
        except Exception as ex:
            msg = 'Failed to visit in {}: {}'.format(self.name, ex)
            logger.debug(msg)
            return
        self.argument_names = visitor.arguments
        self.argument_types = visitor.types
        if function_type != FunctionType.FUNCTION and len(self.argument_names) > 0:
            if not _has_decorator(function, "staticmethod"):
                self.argument_names.pop(0)
                self.argument_types.pop(0)
        self.has_return = bool(visitor.returns)
        self.has_empty_return = False
        if self.has_return:
            return_value = visitor.returns[0]
            self.has_empty_return = (
                return_value is not None
                and return_value.value is None
            )
        self.return_type = _get_return_type(function)
        self.has_yield = bool(visitor.yields)
        self.raises = visitor.exceptions
        self.docstring = _get_docstring(function)
        self.variables = [x.id for x in visitor.variables]
        self.raises_assert = bool(visitor.asserts)
        self.is_abstract = visitor.is_abstract


def get_function_descriptions(program):
    # type: (ast.AST) -> List[FunctionDescription]
    """Get function name, args, return presence and docstrings.

    This function should be called on the top level of the
    document (for functions), and on classes (for methods.)

    Args:
        program: The tree representing the entire program.

    Returns:
        A list of function descriptions pulled from the ast.

    """
    ret = list()  # type: List[FunctionDescription]

    visitor = FunctionAndMethodVisitor()
    visitor.visit(program)
    for prop in visitor.properties:
        ret.append(
            FunctionDescription(function_type=FunctionType.PROPERTY, function=prop)
        )

    for method in visitor.methods:
        ret.append(
            FunctionDescription(function_type=FunctionType.METHOD, function=method)
        )

    for function in visitor.functions:
        ret.append(
            FunctionDescription(function_type=FunctionType.FUNCTION, function=function)
        )

    return ret



import enum


class DocstringStyle(enum.Enum):
    GOOGLE = 0
    SPHINX = 1
    NUMPY = 2

    @classmethod
    def from_string(cls, style):
        style = style.lower().strip()
        if style == 'google':
            return cls.GOOGLE
        if style == 'sphinx':
            return cls.SPHINX
        if style == 'numpy':
            return cls.NUMPY

        raise Exception(
                'Unrecognized style "{}".  Should be one of {}'.format(
                    style,
                    [x.name for x in DocstringStyle]
                )
            )



"""A module to read/define configuration and logging.

This module contains two global instances which should
be accessed only through their getter/setter methods.
These instances are not threadsafe: they should be
updated only prior to spawning any threads.

"""

import configparser
from enum import Enum
import logging
from logging import (  # noqa
    Logger,
)
import os

from typing import (  # noqa
    Iterable,
    List,
    Optional,
)

from .docstring.style import DocstringStyle
from .strictness import Strictness


def get_logger():  # type: () -> Logger
    """Get the default logger for darglint.

    Returns:
        The default logger for darglint.

    """
    return logging.getLogger('darglint')


POSSIBLE_CONFIG_FILENAMES = (
    '.darglint',
    'setup.cfg',
    'tox.ini',
)

DEFAULT_DISABLED = {'DAR104'}


class AssertStyle(Enum):
    """Describes how to handle assertions."""
    RAISE = 1
    LOG = 2


class LogLevel(Enum):
    """Describes the level of error which should be logged.

    These levels correspond to the levels in logging.
    This wrapper primarily serves as a means of conveniently
    parsing the levels, while maintaining the same interface
    as other options.

    """
    CRITICAL = logging.CRITICAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    INFO = logging.INFO
    DEBUG = logging.DEBUG

    @classmethod
    def from_string(cls, level):
        # type: (str) -> LogLevel
        normalized_level = level.lower().strip()
        if normalized_level == 'critical':
            return cls.CRITICAL
        elif normalized_level == 'error':
            return cls.ERROR
        elif normalized_level == 'warning':
            return cls.WARNING
        elif normalized_level == 'info':
            return cls.INFO
        elif normalized_level == 'debug':
            return cls.DEBUG
        else:
            raise ValueError('Unrecognized log level, "{}"'.format(
                level
            ))


class Configuration(object):

    def __init__(self, ignore, message_template, style, strictness,
                 ignore_regex=None, ignore_raise=[], ignore_properties=False, enable=[],
                 indentation=4, assert_style=AssertStyle.LOG,
                 log_level=LogLevel.CRITICAL):
        # type: (List[str], Optional[str], DocstringStyle, Strictness, Optional[str], List[str], bool, List[str], int, AssertStyle, LogLevel) -> None  # noqa: E501
        """Initialize the configuration object.

        Args:
            ignore: A list of error codes to ignore.
            message_template: the template with which to format the errors.
            style: The style of docstring.
            strictness: The minimum strictness to allow.
            ignore_regex: A regular expression which enables ignoring
                functions/methods by name.
            ignore_raise: A list of exceptions that don't need to be
                documented.
            ignore_properties: Bool indicating whether to ignore properties functions
                or not.
            enable: A list of of error codes that are disabled by default.
            indentation: The number of spaces to count as an indent.
            assert_style: The assert style to use (e.g. log on failed
                assertions, or raise exception on failed assertions.)

        """
        self._enable = enable
        self._ignore = ignore
        self.message_template = message_template
        self.style = style
        self.strictness = strictness
        self.errors_to_ignore = self._get_errors_to_ignore()
        self.ignore_regex = ignore_regex
        self.ignore_raise = ignore_raise
        self.ignore_properties = ignore_properties
        self.indentation = indentation
        self.assert_style = assert_style
        self.log_level = log_level

    @property
    def log_level(self):
        # type: () -> LogLevel
        return self._log_level

    @log_level.setter
    def log_level(self, log_level):
        # type: (LogLevel) -> None
        self._log_level = log_level
        logger = get_logger()
        logger.setLevel(log_level.value)

    def __str__(self):
        # type: () -> str
        return '\n'.join([
            'message_template={message_template}',
            'style={style}',
            'strictness={strictness}',
            'indentation={indentation}',
            'ignore={errors_to_ignore}',
            'ignore_regex={ignore_regex}',
            'ignore_raise={ignore_raise}',
        ]).format(**self.__dict__)

    @classmethod
    def get_default_instance(cls):
        return cls(
            ignore=list(),
            message_template=None,
            style=DocstringStyle.GOOGLE,
            strictness=Strictness.FULL_DESCRIPTION,
        )

    @property
    def enable(self):
        # type: () -> List[str]
        return self._enable

    @enable.setter
    def enable(self, errors):
        # type: (List[str]) -> None
        self._enable = errors
        self.errors_to_ignore = self._get_errors_to_ignore()

    @property
    def ignore(self):
        # type: () -> List[str]
        return self._ignore

    @ignore.setter
    def ignore(self, errors):
        # type: (List[str]) -> None
        self._ignore = errors
        self.errors_to_ignore = self._get_errors_to_ignore()

    def _get_errors_to_ignore(self):
        # type: () -> List[str]
        """Update the errors to ignore, accounding for defaults.

        For use in constructing a cached `errors_to_ignore` value.
        Since this value could be used frequently, it makes
        sense to cache this value.

        Returns:
            The errors to ignore, including the default errors.

        """
        disabled = DEFAULT_DISABLED - set(self._enable)
        return self._ignore + list(disabled)


def load_config_file(filename):  # type: (str) -> Configuration
    """Load the config file located at the filename.

    Args:
        filename: A valid filename to read from.

    Raises:
        Exception: When the configuration style is not a valid choice.

    Returns:
        A Configuration object.

    """
    config = configparser.ConfigParser()
    config.read(filename)
    ignore = list()
    enable = list()
    message_template = None
    ignore_regex = None
    ignore_raise = list()
    ignore_properties = False
    style = DocstringStyle.GOOGLE
    strictness = Strictness.FULL_DESCRIPTION
    indentation = 4
    log_level = LogLevel.CRITICAL
    if 'darglint' in config.sections():
        if 'ignore' in config['darglint']:
            errors = config['darglint']['ignore']
            for error in errors.split(','):
                ignore.append(error.strip())
        if 'enable' in config['darglint']:
            to_enable = config['darglint']['enable']
            for error in to_enable.split(','):
                enable.append(error.strip())
        if 'message_template' in config['darglint']:
            message_template = config['darglint']['message_template']
        if 'ignore_regex' in config['darglint']:
            ignore_regex = config['darglint']['ignore_regex']
        if 'ignore_raise' in config['darglint']:
            to_ignore_raise = config['darglint']['ignore_raise']
            for exception in to_ignore_raise.split(','):
                ignore_raise.append(exception.strip())
        if 'ignore_properties' in config['darglint']:
            ignore_properties = bool(config['darglint']['ignore_properties'])
        if 'docstring_style' in config['darglint']:
            raw_style = config['darglint']['docstring_style']
            style = DocstringStyle.from_string(raw_style)

        if 'strictness' in config['darglint']:
            raw_strictness = config['darglint']['strictness']
            strictness = Strictness.from_string(raw_strictness)

        if 'indentation' in config['darglint']:
            try:
                indentation = int(config['darglint']['indentation'])
            except ValueError:
                raise Exception(
                    'Unrecognized value for indentation.  Expected '
                    'a non-zero, positive integer, but received {}'.format(
                        config['darglint']['indentation']
                    )
                )

        if 'log_level' in config['darglint']:
            log_level = LogLevel.from_string(config['darglint']['log_level'])
    return Configuration(
        ignore=ignore,
        message_template=message_template,
        style=style,
        strictness=strictness,
        ignore_regex=ignore_regex,
        ignore_raise=ignore_raise,
        ignore_properties=ignore_properties,
        enable=enable,
        indentation=indentation,
    )


def walk_path():  # type: () -> Iterable[str]
    """Yield directories from the current to root.

    Yields:
        The current directory, then its parent, etc. all
        the way up to root.

    """
    cwd = os.getcwd()
    yield cwd
    prev = cwd
    next_path = os.path.dirname(cwd)

    # Assumes that os.path.dirname will give the root path back
    # when given the root path.
    while prev != next_path:
        yield next_path
        prev = next_path
        next_path = os.path.dirname(next_path)


def find_config_file_in_path(path):  # type: (str) -> Optional[str]
    """Return the config path, if it is correct, or None.

    Args:
        path: The path to check.

    Returns:
        The fully qualified path to the config file, if it is
        in this directory, otherwise none.

    """
    try:
        filenames = os.listdir(path)
    except PermissionError:
        return None
    for filename in filenames:
        if filename in POSSIBLE_CONFIG_FILENAMES:
            config = configparser.ConfigParser()
            fully_qualified_path = os.path.join(path, filename)
            try:
                config.read(fully_qualified_path)
                if 'darglint' in config.sections():
                    return fully_qualified_path
            except configparser.ParsingError:
                get_logger().error('Unable to parse file {}'.format(
                    fully_qualified_path
                ))
    return None


def find_config_file():  # type: () -> Optional[str]
    """Return the location of the config file.

    Returns:
        The location of the config file, if it exists.
        Otherwise, returns None.

    """
    # Check the current directory
    for path in walk_path():
        possible_config_filename = find_config_file_in_path(path)
        if possible_config_filename is not None:
            return possible_config_filename
    return None


def get_config_from_file():  # type: () -> Configuration
    """Locate the configuration file and return its Configuration.

    Returns:
        The Configuration described in the nearest configuration file,
        otherwise an empty Configuration.

    """
    filename = find_config_file()
    if filename is None:
        return Configuration.get_default_instance()
    return load_config_file(filename)


# The global instance of the config file to use.
_config = get_config_from_file()


def get_config():
    """Get the global instance of the configuration.

    This instance is not threadsafe, and should only
    be updated in the initial launching script.
    I considered adding a mutability option, to make
    this more explicit, but I think it's obvious enough
    that you shouldn't modify the configuration elsewhere.

    Returns:
        A global configuration instance.

    """
    return _config



# Generated on 2020-07-25 08:37:49.351851

from darglint.parse.grammar import (
    BaseGrammar,
    P,
)

from darglint.errors import (
    IndentError,
)

from darglint.token import (
    TokenType,
)

from darglint.parse.identifiers import (
    NoqaIdentifier,
)

from darglint.errors import (
    EmptyDescriptionError,
)

class VariableTypeGrammar(BaseGrammar):
    productions = [
        P("variable-type-section", ([], "varhead", "variable-type-section1", 0), ([], "varhead-no-follow", "newlines", 0), ([EmptyDescriptionError], "colon", "varhead-no-follow1", 0)),
        P("varhead", ([], "colon", "varhead0", 0)),
        P("varhead-no-follow", ([EmptyDescriptionError], "colon", "varhead-no-follow1", 0)),
        P("item-body", ([], "line", "item-body0", 2), ([], "line", "item-body1", 2), ([], "line", "item-body2", 2), ([], "word", "line", 2), ([], "word", "noqa-maybe", 2), ([NoqaIdentifier], "hash", "noqa", 2), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 2), (TokenType.INDENT, 2), (TokenType.COLON, 2), (TokenType.HASH, 2), (TokenType.LPAREN, 2), (TokenType.RPAREN, 2), (TokenType.WORD, 2), (TokenType.RAISES, 2), (TokenType.ARGUMENTS, 2), (TokenType.ARGUMENT_TYPE, 2), (TokenType.RETURNS, 2), (TokenType.RETURN_TYPE, 2), (TokenType.YIELDS, 2), (TokenType.YIELD_TYPE, 2), (TokenType.VARIABLES, 2), (TokenType.VARIABLE_TYPE, 2), (TokenType.NOQA, 2), (TokenType.OTHER, 2), (TokenType.RECEIVES, 2), (TokenType.WARNS, 2), (TokenType.SEE, 2), (TokenType.ALSO, 2), (TokenType.NOTES, 2), (TokenType.EXAMPLES, 2), (TokenType.REFERENCES, 2), (TokenType.HEADER, 2), ([IndentError], "line", "item-body7", 1)),
        P("block-indented", ([], "paragraph-indented", "block-indented0", 0), ([], "paragraph-indented", "block-indented1", 0), ([], "indented", "paragraph-indented0", 0), ([], "indented", "line", 0)),
        P("split-indented", ([], "newline", "split-indented0", 0), (TokenType.NEWLINE, 0)),
        P("paragraph-indented", ([], "indented", "paragraph-indented0", 0), ([], "indented", "line", 0)),
        P("indented", ([], "indent", "indents", 0), (TokenType.INDENT, 0)),
        P("block", ([], "paragraph", "block0", 0), ([], "indents", "paragraph0", 0), ([], "indents", "line", 0), ([], "line", "paragraph2", 0), ([], "word", "line", 0), ([], "word", "noqa-maybe", 0), ([NoqaIdentifier], "hash", "noqa", 0), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 0), (TokenType.INDENT, 0), (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0), ([], "line", "paragraph1", 0)),
        P("paragraph", ([], "indents", "paragraph0", 0), ([], "indents", "line", 0), ([], "line", "paragraph2", 0), ([], "word", "line", 0), ([], "word", "noqa-maybe", 0), ([NoqaIdentifier], "hash", "noqa", 0), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 0), (TokenType.INDENT, 0), (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0), ([], "line", "paragraph1", 0)),
        P("line", ([], "word", "line", 0), ([], "word", "noqa-maybe", 0), ([NoqaIdentifier], "hash", "noqa", 0), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 0), (TokenType.INDENT, 0), (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0)),
        P("indents", ([], "indent", "indents", 0), (TokenType.INDENT, 0)),
        P("split", ([], "newline", "split0", 0)),
        P("newlines", ([], "newline", "newlines", 0), (TokenType.NEWLINE, 0)),
        P("word", (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.INDENT, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0)),
        P("colon", (TokenType.COLON, 0)),
        P("hash", (TokenType.HASH, 0)),
        P("indent", (TokenType.INDENT, 0)),
        P("newline", (TokenType.NEWLINE, 0)),
        P("variable-type", (TokenType.VARIABLE_TYPE, 0)),
        P("noqa", (TokenType.NOQA, 0)),
        P("noqa-maybe", ([NoqaIdentifier], "hash", "noqa", 0), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 0)),
        P("noqa-head", ([], "hash", "noqa", 0)),
        P("words", ([], "word", "words", 0), (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.INDENT, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0)),
        P("variable-type-section1", ([], "item-body", "newlines", 0), ([], "line", "item-body0", 2), ([], "line", "item-body1", 2), ([], "line", "item-body2", 2), ([], "word", "line", 2), ([], "word", "noqa-maybe", 2), ([NoqaIdentifier], "hash", "noqa", 2), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 2), (TokenType.INDENT, 2), (TokenType.COLON, 2), (TokenType.HASH, 2), (TokenType.LPAREN, 2), (TokenType.RPAREN, 2), (TokenType.WORD, 2), (TokenType.RAISES, 2), (TokenType.ARGUMENTS, 2), (TokenType.ARGUMENT_TYPE, 2), (TokenType.RETURNS, 2), (TokenType.RETURN_TYPE, 2), (TokenType.YIELDS, 2), (TokenType.YIELD_TYPE, 2), (TokenType.VARIABLES, 2), (TokenType.VARIABLE_TYPE, 2), (TokenType.NOQA, 2), (TokenType.OTHER, 2), (TokenType.RECEIVES, 2), (TokenType.WARNS, 2), (TokenType.SEE, 2), (TokenType.ALSO, 2), (TokenType.NOTES, 2), (TokenType.EXAMPLES, 2), (TokenType.REFERENCES, 2), (TokenType.HEADER, 2), ([IndentError], "line", "item-body7", 1)),
        P("varhead0", ([], "variable-type", "varhead1", 0)),
        P("varhead1", ([], "word", "colon", 0)),
        P("varhead-no-follow1", ([], "variable-type", "varhead-no-follow2", 0)),
        P("varhead-no-follow2", ([], "word", "colon", 0)),
        P("item-body0", ([], "newline", "block-indented", 0)),
        P("item-body1", ([], "newlines", "block-indented", 0), ([], "paragraph-indented", "block-indented0", 0), ([], "paragraph-indented", "block-indented1", 0), ([], "indented", "paragraph-indented0", 0), ([], "indented", "line", 0)),
        P("item-body2", ([], "newline", "item-body3", 0)),
        P("item-body3", ([], "indent", "item-body4", 0)),
        P("item-body4", ([], "newline", "item-body5", 0)),
        P("item-body5", ([], "newlines", "block-indented", 0), ([], "paragraph-indented", "block-indented0", 0), ([], "paragraph-indented", "block-indented1", 0), ([], "indented", "paragraph-indented0", 0), ([], "indented", "line", 0)),
        P("item-body7", ([], "newline", "block", 0)),
        P("block-indented0", ([], "split", "block-indented", 0)),
        P("block-indented1", ([], "split-indented", "block-indented", 0)),
        P("split-indented0", ([], "indents", "newlines", 0), ([], "newline", "newlines", 0), (TokenType.NEWLINE, 0), ([], "indent", "indents", 0), (TokenType.INDENT, 0)),
        P("paragraph-indented0", ([], "line", "paragraph-indented1", 0)),
        P("paragraph-indented1", ([], "newline", "paragraph-indented", 0)),
        P("block0", ([], "split", "block", 0)),
        P("paragraph0", ([], "line", "paragraph1", 0)),
        P("paragraph1", ([], "newline", "paragraph", 0)),
        P("paragraph2", ([], "newline", "paragraph", 0)),
        P("split0", ([], "newline", "newlines", 0), (TokenType.NEWLINE, 0)),
        P("noqa-statement1", ([], "colon", "words", 0)),
    ]
    start = "variable-type-section"


"""Defines the command line interface for darglint."""
import argparse
import ast
import pathlib
import sys
import inspect

from typing import (
    List,
)

from .function_description import (
    read_program,
    get_function_descriptions,
)
from .integrity_checker import IntegrityChecker
from .config import (
    get_config,
    get_logger,
    LogLevel,
)
from .docstring.style import DocstringStyle
from .strictness import Strictness
import darglint.errors
from darglint.error_report import ErrorReport


# ---------------------- ARGUMENT PARSER -----------------------------

parser = argparse.ArgumentParser(description='Check docstring validity.')
parser.add_argument(
    '--message-template',
    '-m',
    type=str,
    help=(
        'Specify a message template.  This is a Python format string '
        'describing errors, which can access the following attributes:\n'
        '    line: The line number,\n'
        '    msg: The error message,\n'
        '    msg_id: The error code,\n'
        '    obj: The function/method name,\n'
        '    path: The relative file path.\n'
    ),
)
parser.add_argument(
    '--ignore-regex',
    '-i',
    type=str,
    help=(
        'Methods/function names matching this regex will be skipped '
        'during analysis.'
    ),
)
parser.add_argument(
    '--ignore-raise',
    '-c',
    type=str,
    help=(
        'Exceptions that don\'t need to be documented in docstrings. '
        'Accepts a comma-separated list. E.g.: "ValueError,MyCustomError"'
    ),
)
parser.add_argument(
    '--ignore-properties',
    '-p',
    action="store_true",
    default=False,
    help='Class property methods will be ignored by darglint',
)
parser.add_argument(
    '--raise-syntax',
    action='store_true',
    help=(
        'When a docstring is incorrectly formatted, raise an exception '
        'rather than storing the error.  Useful for debugging darglint.'
    ),
)
parser.add_argument(
    '--verbosity',
    '-v',
    default=1,
    type=int,
    choices=[1, 2],
    help='The level of verbosity.',
)
parser.add_argument(
    '--version',
    action='store_true',
    help=(
        'Return the current version number of darglint.'
    ),
)
parser.add_argument(
    'files',
    nargs='*',
    help=(
        'The python source files to check. If "-" is given, then stdin will '
        'be read.'
    ),
)
parser.add_argument(
    '--no-exit-code',
    '-x',
    action='store_true',
    help=(
        'Exit with status 0, even on errors.  By default, darglint '
        'exits with status 1 when errors are encountered.  Giving '
        'this flag prevents that.  Useful when invocating with xargs '
        'and you want to see all errors.  '
        'Ex: `find . -name "*.py" | xargs darglint -x`'
    ),
)
parser.add_argument(
    '--list-errors',
    action='store_true',
    help=(
        'Print a list of error codes and what they represent.'
    )
)
parser.add_argument(
    '-s',
    '--docstring-style',
    default=None,
    choices=['google', 'sphinx', 'numpy'],
    help=(
        'The docstring style used in the given project. Currently, '
        'only google, sphinx, and numpy styles are supported.'
    )
)
parser.add_argument(
    '-z',
    '--strictness',
    default=None,
    choices=[
        'short',
        'long',
        'full',
    ],
    help=(
        'The minimum strictness when checking docstrings. '
        '`short`, for example, will result in one-line '
        'docstrings always being accepted.  Anything more than one line '
        'would go through the full check.'
    ),
)
parser.add_argument(
    '-e',
    '--enable',
    type=str,
    help=(
        'Enable disabled-by-default errors.  Accepts a '
        'comma-separated list of error codes.  E.g.: '
        '"DAR104,DAR105"'
    )
)
parser.add_argument(
    '--indentation',
    type=int,
    default=None,
    help=(
        'The number of spaces to count as an indentation. '
        'For example, if following the Google python style '
        'guide, you would set --indentation=2.'
    ),
)
parser.add_argument(
    '--log-level',
    '-l',
    type=str,
    default=None,
    choices=[
        'CRITICAL',
        'ERROR',
        'WARNING',
        'INFO',
        'DEBUG',
    ],
    help=(
        'The level at which to log.  Can help with debugging '
        'when something strange is happening.  The default '
        'level, CRITICAL, means that only the most severe of '
        'errors will be logged.  Assertions are logged at the '
        'ERROR level.'
    )
)

# ---------------------- MAIN SCRIPT ---------------------------------


def get_error_report(filename,
                     verbosity,
                     raise_errors_for_syntax,
                     message_template=None,
                     ):
    # type: (str, int, bool, str) -> str
    """Get the error report for the given file.

    Args:
        filename: The name of the module to check.
        verbosity: The level of verbosity, in the range [1, 3].
        raise_errors_for_syntax: True if we want parser errors
            to propagate up (crashing darglint.)  This is useful
            if we are developing on darglint -- we can get the stack
            trace and know exactly where darglint failed.
        message_template: A python format string for specifying
            how the message should appear to the user.

    Returns:
        An error report for the file.

    """
    program = read_program(filename)
    try:
        tree = ast.parse(program)
        functions = get_function_descriptions(tree)
        checker = IntegrityChecker(
            raise_errors=raise_errors_for_syntax,
        )
        for function in functions:
            checker.schedule(function)
        return checker.get_error_report_string(
            verbosity,
            filename,
            message_template=message_template,
        )
    except SyntaxError as e:
        error = darglint.errors.PythonSyntaxError(e)
        report = ErrorReport([error], filename, verbosity, message_template)
        return str(report)


def print_error_list():
    errors = list()  # type: List[str]
    for name, obj in inspect.getmembers(darglint.errors, inspect.isclass):
        if (issubclass(obj, darglint.errors.DarglintError)
                and obj != darglint.errors.DarglintError):
            errors.append('{}: {}'.format(obj.error_code, obj.description))
    errors.sort()
    print('\n'.join(errors))


def print_version():
    print('1.8.0')


def main():
    # type: () -> None
    """Run darglint.

    Called as a script when setup.py is installed.

    """
    args = parser.parse_args()
    exit_code = not args.no_exit_code
    encountered_errors = False

    if args.list_errors:
        print_error_list()
        sys.exit(0)

    if args.version:
        print_version()

    # Expand directories.
    files = []
    for f in args.files:
        p = pathlib.Path(f)
        if not p.is_dir() and p.suffix == '.py':
            files.append(f)
        # Convert back to strings to not require modifications of any
        # subsequent code.
        files.extend(str(i) for i in p.glob('**/*.py'))

    try:
        config = get_config()

        # Only override enable if explicitly passed.
        if args.enable:
            config.enable = [
                x.strip() for x in args.enable.split(',')
            ]

        if '*' in config.ignore:
            sys.exit(0)

        if args.indentation:
            config.indentation = args.indentation

        if args.docstring_style == 'sphinx':
            config.style = DocstringStyle.SPHINX
        elif args.docstring_style == 'google':
            config.style = DocstringStyle.GOOGLE
        elif args.docstring_style == 'numpy':
            config.style = DocstringStyle.NUMPY

        if args.strictness == 'short':
            config.strictness = Strictness.SHORT_DESCRIPTION
        elif args.strictness == 'long':
            config.strictness = Strictness.LONG_DESCRIPTION
        elif args.strictness == 'full':
            config.strictness = Strictness.FULL_DESCRIPTION

        if args.log_level:
            config.log_level = LogLevel.from_string(args.log_level)

        if args.ignore_regex:
            config.ignore_regex = args.ignore_regex
        if args.ignore_raise:
            config.ignore_raise = [
                x.strip() for x in args.ignore_raise.split(",")
            ]
        if args.ignore_properties:
            config.ignore_properties = args.ignore_properties

        raise_errors_for_syntax = args.raise_syntax or False
        for filename in files:
            error_report = get_error_report(
                filename,
                args.verbosity,
                raise_errors_for_syntax,
                message_template=args.message_template,
            )
            if error_report:
                print(error_report + '\n')
                encountered_errors = True
    except Exception as exc:
        # Exit with status 129 regardless of whether user wants a
        # exit code or not -- darglint failed, and it should
        # look like it failed.
        logger = get_logger()
        logger.critical(exc)
        sys.exit(129)
    if encountered_errors and exit_code:
        sys.exit(1)
    else:
        sys.exit(0)


if __name__ == '__main__':
    main()



# Generated on 2020-04-04 11:23:41.567044

from darglint.parse.grammar import (
    BaseGrammar,
    P,
)

from darglint.token import (
    TokenType,
)

from darglint.parse.identifiers import (
    NoqaIdentifier,
)

class YieldsWithoutTypeGrammar(BaseGrammar):
    productions = [
        P("yields-section", ([], "yields-head", "yields-body", 0)),
        P("yields-head", ([], "yields", "yields-head0", 0)),
        P("yields-body", ([], "indented", "yields-content", 0)),
        P("yields-content", ([], "line", "newlines", 0), ([], "line", "yields-content0", 0), ([], "word", "line", 0), ([], "word", "noqa-maybe", 0), ([NoqaIdentifier], "hash", "noqa", 0), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 0), (TokenType.INDENT, 0), (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0)),
        P("block-indented", ([], "paragraph-indented", "block-indented0", 0), ([], "indented", "paragraph-indented0", 0), ([], "indented", "line", 0)),
        P("paragraph-indented", ([], "indented", "paragraph-indented0", 0), ([], "indented", "line", 0)),
        P("line", ([], "word", "line", 0), ([], "word", "noqa-maybe", 0), ([NoqaIdentifier], "hash", "noqa", 0), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 0), (TokenType.INDENT, 0), (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0)),
        P("indented", ([], "indent", "indents", 0), (TokenType.INDENT, 0)),
        P("indents", ([], "indent", "indents", 0), (TokenType.INDENT, 0)),
        P("split", ([], "newline", "split0", 0)),
        P("newlines", ([], "newline", "newlines", 0), (TokenType.NEWLINE, 0)),
        P("word", (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.INDENT, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0)),
        P("colon", (TokenType.COLON, 0)),
        P("hash", (TokenType.HASH, 0)),
        P("indent", (TokenType.INDENT, 0)),
        P("newline", (TokenType.NEWLINE, 0)),
        P("yields", (TokenType.YIELDS, 0)),
        P("noqa", (TokenType.NOQA, 0)),
        P("noqa-maybe", ([NoqaIdentifier], "hash", "noqa", 0), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 0)),
        P("noqa-head", ([], "hash", "noqa", 0)),
        P("words", ([], "word", "words", 0), (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.INDENT, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0)),
        P("yields-head0", ([], "colon", "newline", 0)),
        P("yields-content0", ([], "newline", "yields-content1", 0)),
        P("yields-content1", ([], "block-indented", "newlines", 0), ([], "paragraph-indented", "block-indented0", 0), ([], "indented", "paragraph-indented0", 0), ([], "indented", "line", 0)),
        P("block-indented0", ([], "split", "block-indented", 0)),
        P("paragraph-indented0", ([], "line", "paragraph-indented1", 0)),
        P("paragraph-indented1", ([], "newline", "paragraph-indented", 0)),
        P("split0", ([], "newline", "newlines", 0), (TokenType.NEWLINE, 0)),
        P("noqa-statement1", ([], "colon", "words", 0)),
    ]
    start = "yields-section"


# Generated on 2020-04-23 21:45:16.239170

from darglint.parse.grammar import (
    BaseGrammar,
    P,
)

from darglint.token import (
    TokenType,
)

from darglint.parse.identifiers import (
    NoqaIdentifier,
)

from darglint.errors import (
    EmptyDescriptionError,
    IndentError,
)

from darglint.parse.identifiers import (
    ExceptionIdentifier,
    ExceptionItemIdentifier,
)

class RaisesGrammar(BaseGrammar):
    productions = [
        P("raises-section", ([], "raises", "raises-section1", 0)),
        P("items-exception", ([], "item-exception", "items-exception0", 0), ([ExceptionItemIdentifier], "head-exception", "item-exception1", 2), ([ExceptionItemIdentifier], "head-exception", "line", 2), ([IndentError, ExceptionItemIdentifier], "head-exception", "item-exception5", 0), ([IndentError, ExceptionItemIdentifier], "head-exception", "item-exception8", 0), ([IndentError, ExceptionItemIdentifier], "head-exception", "item-exception12", 0), ([ExceptionIdentifier, EmptyDescriptionError], "indent", "head-exception1", 0)),
        P("item-exception", ([ExceptionItemIdentifier], "head-exception", "item-exception1", 2), ([ExceptionItemIdentifier], "head-exception", "line", 2), ([IndentError, ExceptionItemIdentifier], "head-exception", "item-exception5", 0), ([IndentError, ExceptionItemIdentifier], "head-exception", "item-exception8", 0), ([IndentError, ExceptionItemIdentifier], "head-exception", "item-exception12", 0), ([ExceptionIdentifier, EmptyDescriptionError], "indent", "head-exception1", 0)),
        P("head-exception", ([ExceptionIdentifier], "indent", "head-exception1", 0)),
        P("exception", (TokenType.WORD, 0)),
        P("paragraph-indented-two", ([], "indented-two", "paragraph-indented-two0", 0), ([], "indented-two", "line", 0)),
        P("paragraph", ([], "indents", "paragraph0", 0), ([], "indents", "line", 0), ([], "line", "paragraph2", 0), ([], "word", "line", 0), ([], "word", "noqa-maybe", 0), ([NoqaIdentifier], "hash", "noqa", 0), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 0), (TokenType.INDENT, 0), (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0), ([], "line", "paragraph1", 0)),
        P("line", ([], "word", "line", 0), ([], "word", "noqa-maybe", 0), ([NoqaIdentifier], "hash", "noqa", 0), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 0), (TokenType.INDENT, 0), (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0)),
        P("indented-two", ([], "indent", "indented-two0", 0)),
        P("indents", ([], "indent", "indents", 0), (TokenType.INDENT, 0)),
        P("newlines", ([], "newline", "newlines", 0), (TokenType.NEWLINE, 0)),
        P("word", (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.INDENT, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0)),
        P("colon", (TokenType.COLON, 0)),
        P("hash", (TokenType.HASH, 0)),
        P("indent", (TokenType.INDENT, 0)),
        P("newline", (TokenType.NEWLINE, 0)),
        P("raises", (TokenType.RAISES, 0)),
        P("noqa", (TokenType.NOQA, 0)),
        P("noqa-maybe", ([NoqaIdentifier], "hash", "noqa", 0), ([NoqaIdentifier], "noqa-head", "noqa-statement1", 0)),
        P("noqa-head", ([], "hash", "noqa", 0)),
        P("words", ([], "word", "words", 0), (TokenType.COLON, 0), (TokenType.HASH, 0), (TokenType.INDENT, 0), (TokenType.LPAREN, 0), (TokenType.RPAREN, 0), (TokenType.WORD, 0), (TokenType.RAISES, 0), (TokenType.ARGUMENTS, 0), (TokenType.ARGUMENT_TYPE, 0), (TokenType.RETURNS, 0), (TokenType.RETURN_TYPE, 0), (TokenType.YIELDS, 0), (TokenType.YIELD_TYPE, 0), (TokenType.VARIABLES, 0), (TokenType.VARIABLE_TYPE, 0), (TokenType.NOQA, 0), (TokenType.OTHER, 0), (TokenType.RECEIVES, 0), (TokenType.WARNS, 0), (TokenType.SEE, 0), (TokenType.ALSO, 0), (TokenType.NOTES, 0), (TokenType.EXAMPLES, 0), (TokenType.REFERENCES, 0), (TokenType.HEADER, 0)),
        P("raises-section1", ([], "colon", "raises-section2", 0)),
        P("raises-section2", ([], "newline", "raises-section3", 0)),
        P("raises-section3", ([], "items-exception", "newlines", 0), ([], "item-exception", "items-exception0", 0), ([ExceptionItemIdentifier], "head-exception", "item-exception1", 2), ([ExceptionItemIdentifier], "head-exception", "line", 2), ([IndentError, ExceptionItemIdentifier], "head-exception", "item-exception5", 0), ([IndentError, ExceptionItemIdentifier], "head-exception", "item-exception8", 0), ([IndentError, ExceptionItemIdentifier], "head-exception", "item-exception12", 0), ([ExceptionIdentifier, EmptyDescriptionError], "indent", "head-exception1", 0)),
        P("items-exception0", ([], "newline", "items-exception", 0)),
        P("item-exception1", ([], "line", "item-exception2", 0)),
        P("item-exception2", ([], "newline", "paragraph-indented-two", 0)),
        P("item-exception5", ([], "line", "item-exception6", 0)),
        P("item-exception6", ([], "newline", "paragraph", 0)),
        P("item-exception8", ([], "line", "item-exception9", 0)),
        P("item-exception9", ([], "newline", "item-exception10", 0)),
        P("item-exception10", ([], "paragraph", "newline", 0)),
        P("item-exception12", ([], "newline", "paragraph", 0)),
        P("head-exception1", ([], "exception", "colon", 0)),
        P("paragraph-indented-two0", ([], "line", "paragraph-indented-two1", 0)),
        P("paragraph-indented-two1", ([], "newline", "paragraph-indented-two", 0)),
        P("paragraph0", ([], "line", "paragraph1", 0)),
        P("paragraph1", ([], "newline", "paragraph", 0)),
        P("paragraph2", ([], "newline", "paragraph", 0)),
        P("indented-two0", ([], "indent", "indents", 0), (TokenType.INDENT, 0)),
        P("noqa-statement1", ([], "colon", "words", 0)),
    ]
    start = "raises-section"


"""Defines IntegrityChecker."""

import re
import concurrent.futures
from typing import (  # noqa: F401
    Any,
    cast,
    List,
    Optional,
    Set,
)

from .function_description import (  # noqa: F401
    FunctionDescription,
)
from .docstring.base import BaseDocstring
from .docstring.docstring import Docstring
from .docstring.sections import Sections
from .docstring.style import DocstringStyle
from .errors import (  # noqa: F401
    DarglintError,
    ExcessParameterError,
    ExcessRaiseError,
    ExcessReturnError,
    ExcessVariableError,
    ExcessYieldError,
    MissingParameterError,
    MissingRaiseError,
    MissingReturnError,
    MissingYieldError,
    ParameterTypeMismatchError,
    ParameterTypeMissingError,
    ReturnTypeMismatchError,
)
from .error_report import (
    ErrorReport,
)
from .config import get_config
from .strictness import Strictness


SYNTAX_NOQA = re.compile(r'#\s*noqa:\sS001')
EXPLICIT_GLOBAL_NOQA = re.compile(r'#\s*noqa:\s*\*')
BARE_NOQA = re.compile(r'#\s*noqa([^:]|$)')


class IntegrityChecker(object):
    """Checks the integrity of the docstring compared to the definition."""

    def __init__(self, raise_errors=False):
        # type: (bool) -> None
        """Create a new checker for the given function and docstring.

        Args:
            raise_errors: If true, we will allow ParserExceptions to
                propagate, crashing darglint.  This is mostly useful
                for development.

        """
        self.errors = list()  # type: List[DarglintError]
        self._sorted = True
        self.config = get_config()
        self.raise_errors = raise_errors

        # TODO: Move max workers into a configuration option.
        # A thread pool for handling checks.  Tasks are added to the
        # pool when `schedule` is executed, if it has a docstring.
        # The pool is collected when `get_error_report_string` is called.
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=4)

    def schedule(self, function):
        # type: (FunctionDescription) -> None
        if self._skip_checks(function):
            return

        self.executor.submit(self.run_checks, function)

    def run_checks(self, function):
        # type: (FunctionDescription) -> None
        """Run checks on the given function.

        Args:
            function: A function whose docstring we are verifying.

        Raises:
            Exception: If the docstring format isn't supported.

        """
        if self._skip_checks(function):
            return

        function_docstring = cast(str, function.docstring)
        if self.config.style == DocstringStyle.GOOGLE:
            docstring = Docstring.from_google(
                function_docstring,
            )
        elif self.config.style == DocstringStyle.SPHINX:
            docstring = Docstring.from_sphinx(
                function_docstring,
            )
            self._check_variables(docstring, function)
        elif self.config.style == DocstringStyle.NUMPY:
            docstring = Docstring.from_numpy(
                function_docstring,
            )
        else:
            raise Exception('Unsupported docstring format.')
        if self.config.strictness != Strictness.FULL_DESCRIPTION:
            if docstring.satisfies_strictness(self.config.strictness):
                return
        if docstring.ignore_all:
            return
        self._check_parameters(docstring, function)
        self._check_parameter_types(docstring, function)
        self._check_parameter_types_missing(docstring, function)
        self._check_return(docstring, function)
        self._check_return_type(docstring, function)
        self._check_yield(docstring, function)
        self._check_raises(docstring, function)
        self._check_style(docstring, function)
        self._sorted = False

    def _skip_checks(self, function):
        # type: (FunctionDescription) -> bool
        no_docsting = function.docstring is None
        skip_by_regex = (
            self.config.ignore_regex and
            re.match(self.config.ignore_regex, function.name)
        )
        skip_property = (
            self.config.ignore_properties and function.is_property
        )

        return bool(no_docsting or skip_by_regex or skip_property)

    def _check_parameter_types(self, docstring, function):
        # type: (BaseDocstring, FunctionDescription) -> None
        error_code = ParameterTypeMismatchError.error_code
        if self._ignore_error(docstring, ParameterTypeMismatchError):
            return

        argument_types = dict(
            zip(docstring.get_items(Sections.ARGUMENTS_SECTION) or [],
                docstring.get_types(Sections.ARGUMENTS_SECTION) or [])
        )
        doc_arg_types = list()  # type: List[Optional[str]]
        for name in function.argument_names:
            if name not in argument_types:
                doc_arg_types.append(None)
            else:
                doc_arg_types.append(argument_types[name])
        noqa_lookup = docstring.get_noqas()
        for name, expected, actual in zip(
                function.argument_names,
                function.argument_types,
                doc_arg_types,
        ):
            if expected is None or actual is None:
                continue
            noqa_exists = error_code in noqa_lookup
            name_has_noqa = noqa_exists and name in noqa_lookup[error_code]
            if not (expected == actual or name_has_noqa):
                default_line_numbers = docstring.get_line_numbers(
                    'arguments-section'
                )
                line_numbers = docstring.get_line_numbers_for_value(
                    'ident',
                    name,
                ) or default_line_numbers
                self.errors.append(
                    ParameterTypeMismatchError(
                        function.function,
                        name=name,
                        expected=expected,
                        actual=actual,
                        line_numbers=line_numbers,
                    )
                )

    def _check_parameter_types_missing(self, docstring, function):
        # type: (BaseDocstring, FunctionDescription) -> None
        error_code = ParameterTypeMissingError.error_code
        if self._ignore_error(docstring, ParameterTypeMissingError):
            return

        argument_types = dict(
            zip(docstring.get_items(Sections.ARGUMENTS_SECTION) or [],
                docstring.get_types(Sections.ARGUMENTS_SECTION) or [])
        )

        noqa_lookup = docstring.get_noqas()
        noqa_exists = error_code in noqa_lookup

        for name, argument_type in argument_types.items():
            name_has_no_qa = noqa_exists and name in noqa_lookup[error_code]

            if argument_type is None and not name_has_no_qa:
                default_line_numbers = docstring.get_line_numbers(
                    'arguments-section'
                )
                line_numbers = docstring.get_line_numbers_for_value(
                    'ident',
                    name,
                ) or default_line_numbers
                self.errors.append(
                    ParameterTypeMissingError(
                        function.function,
                        name=name,
                        line_numbers=line_numbers,
                    )
                )

    def _check_return_type(self, docstring, function):
        # type: (BaseDocstring, FunctionDescription) -> None
        if function.is_abstract:
            return

        if self._ignore_error(docstring, ReturnTypeMismatchError):
            return

        fun_type = function.return_type
        doc_type = docstring.get_types(Sections.RETURNS_SECTION)
        if not doc_type or isinstance(doc_type, list):
            doc_type = None
        if fun_type is not None and doc_type is not None:
            if fun_type != doc_type:
                line_numbers = docstring.get_line_numbers(
                    'returns-section',
                )
                self.errors.append(
                    ReturnTypeMismatchError(
                        function.function,
                        expected=fun_type,
                        actual=doc_type,
                        line_numbers=line_numbers,
                    ),
                )

    def _check_yield(self, docstring, function):
        # type: (BaseDocstring, FunctionDescription) -> None
        if function.is_abstract:
            return

        doc_yield = docstring.get_section(Sections.YIELDS_SECTION)
        fun_yield = function.has_yield
        ignore_missing = self._ignore_error(docstring, MissingYieldError)
        ignore_excess = self._ignore_error(docstring, ExcessYieldError)
        if fun_yield and not doc_yield and not ignore_missing:
            self.errors.append(
                MissingYieldError(function.function)
            )
        elif doc_yield and not fun_yield and not ignore_excess:
            line_numbers = docstring.get_line_numbers(
                'yields-section',
            )
            self.errors.append(
                ExcessYieldError(
                    function.function,
                    line_numbers=line_numbers,
                )
            )

    def _check_return(self, docstring, function):
        # type: (BaseDocstring, FunctionDescription) -> None

        if function.is_abstract:
            return

        # If there is an empty return, we don't want to make any
        # judgement about whether it should be reported, as it is
        # probably an early escape.
        if function.has_empty_return:
            return

        doc_return = docstring.get_section(Sections.RETURNS_SECTION)
        fun_return = function.has_return
        ignore_missing = self._ignore_error(docstring, MissingReturnError)
        ignore_excess = self._ignore_error(docstring, ExcessReturnError)
        if fun_return and not doc_return and not ignore_missing:
            self.errors.append(
                MissingReturnError(function.function)
            )
        elif doc_return and not fun_return and not ignore_excess:
            line_numbers = docstring.get_line_numbers(
                'returns-section',
            )
            self.errors.append(
                ExcessReturnError(
                    function.function,
                    line_numbers=line_numbers,
                )
            )

    def _check_parameters(self, docstring, function):
        # type: (BaseDocstring, FunctionDescription) -> None
        docstring_arguments = set(docstring.get_items(
            Sections.ARGUMENTS_SECTION
        ) or [])
        actual_arguments = set(function.argument_names)
        missing_in_doc = actual_arguments - docstring_arguments
        missing_in_doc = self._remove_ignored(
            docstring,
            missing_in_doc,
            MissingParameterError,
        )

        # Get a default line number.
        default_line_numbers = docstring.get_line_numbers(
            'arguments-section'
        )

        for missing in missing_in_doc:
            # See if the documented argument begins with one
            # or two asterisks.
            if (
                (missing.startswith('**')
                    and missing[2:] in docstring_arguments)
                or (missing.startswith('*')
                    and missing[1:] in docstring_arguments)
            ):
                continue

            # Don't require private arguments.
            if missing.startswith('_'):
                continue

            # We use the default line numbers because a missing
            # parameter, by definition, will not have line numbers.
            self.errors.append(
                MissingParameterError(
                    function.function,
                    missing,
                    line_numbers=default_line_numbers
                )
            )

        missing_in_function = docstring_arguments - actual_arguments
        missing_in_function = self._remove_ignored(
            docstring,
            missing_in_function,
            ExcessParameterError,
        )
        for missing in missing_in_function:
            # If the actual argument begins with asterisk(s),
            # then check to make sure the unasterisked version
            # is not missing.
            if (
                '*' + missing in actual_arguments or
                '**' + missing in actual_arguments
            ):
                continue
            line_numbers = docstring.get_line_numbers_for_value(
                'arguments-section',
                missing,
            ) or default_line_numbers
            self.errors.append(
                ExcessParameterError(
                    function.function,
                    missing,
                    line_numbers=line_numbers,
                )
            )

    def _check_variables(self, docstring, function):
        # type: (BaseDocstring, FunctionDescription) -> None
        described_variables = set(
            docstring.get_items(Sections.VARIABLES_SECTION) or []
        )  # type: Set[str]
        actual_variables = set(function.variables)
        excess_in_doc = described_variables - actual_variables

        # Get a default line number.
        default_line_numbers = docstring.get_line_numbers(
            'variables-section',
        )

        for excess in excess_in_doc:
            line_numbers = docstring.get_line_numbers_for_value(
                'variables-section',
                excess,
            ) or default_line_numbers
            self.errors.append(
                ExcessVariableError(
                    function.function,
                    excess,
                    line_numbers=line_numbers,
                )
            )

    def _ignore_error(self, docstring, error):
        # type: (BaseDocstring, Any) -> bool
        """Return true if we should ignore this error.

        Args:
            docstring: The docstring we are reporting on.
            error: The error we might be ignoring.

        Returns:
            True if we should ignore all instances of this error,
            otherwise false.

        """
        error_code = error.error_code
        if error_code in self.config.errors_to_ignore:
            return True
        noqa_lookup = docstring.get_noqas()
        inline_error = error_code in noqa_lookup
        if inline_error and not noqa_lookup[error_code]:
            return True
        return False

    def _remove_ignored(self, docstring, missing, error):
        # type: (BaseDocstring, Set[str], Any) -> Set[str]
        """Remove ignored from missing.

        Args:
            docstring: The docstring we are reporting on.
            missing: A set of missing items.
            error: The error being checked.

        Returns:
            A set of missing items without those to be ignored.

        """
        error_code = error.error_code

        # Ignore globally
        if self._ignore_error(docstring, error):
            return set()

        # There are no noqa statements
        noqa_lookup = docstring.get_noqas()
        inline_ignore = error_code in noqa_lookup
        if not inline_ignore:
            return missing

        # We are to ignore specific instances.
        return missing - set(noqa_lookup[error_code])

    def _check_style(self, docstring, function):
        # type: (BaseDocstring, FunctionDescription) -> None
        for StyleError, line_numbers in docstring.get_style_errors():
            if self._ignore_error(docstring, StyleError):
                continue
            self.errors.append(StyleError(
                function.function,
                line_numbers,
            ))

    def _check_raises(self, docstring, function):
        # type: (BaseDocstring, FunctionDescription) -> None
        if function.is_abstract:
            return

        exception_types = docstring.get_items(Sections.RAISES_SECTION)
        docstring_raises = set(exception_types or [])
        actual_raises = function.raises
        ignore_raise = set(self.config.ignore_raise)
        missing_in_doc = actual_raises - docstring_raises - ignore_raise

        missing_in_doc = self._remove_ignored(
            docstring,
            missing_in_doc,
            MissingRaiseError,
        )

        for missing in missing_in_doc:
            self.errors.append(
                MissingRaiseError(function.function, missing)
            )

        # TODO: Disable by default.
        #
        # Should we even include this?  It seems like the user
        # would know if this function would be likely to raise
        # a certain exception from underlying calls.
        #
        missing_in_function = docstring_raises - actual_raises
        missing_in_function = self._remove_ignored(
            docstring,
            missing_in_function,
            ExcessRaiseError,
        )

        # Remove AssertionError if there is an assert.
        if 'AssertionError' in missing_in_function:
            if function.raises_assert:
                missing_in_function.remove('AssertionError')

        default_line_numbers = docstring.get_line_numbers(
            'raises-section',
        )
        for missing in missing_in_function:
            line_numbers = docstring.get_line_numbers_for_value(
                'raises-section',
                missing,
            ) or default_line_numbers
            self.errors.append(
                ExcessRaiseError(
                    function.function,
                    missing,
                    line_numbers=line_numbers,
                )
            )

    def _sort(self):
        # type: () -> None
        if not self._sorted:
            self.errors.sort(key=lambda x: x.function.lineno)
            self._sorted = True

    def get_error_report(self, verbosity, filename, message_template=None):
        # type: (int, str, str) -> ErrorReport
        self.executor.shutdown()
        return ErrorReport(
            errors=self.errors,
            filename=filename,
            verbosity=verbosity,
            message_template=message_template or self.config.message_template,
        )

    def get_error_report_string(self,
                                verbosity,
                                filename,
                                message_template=None):
        # type: (int, str, str) -> str
        """Return a string representation of the errors.

        Args:
            verbosity: The level of verbosity.  Should be an integer
                in the range [1,3].
            filename: The filename of where the error occurred.
            message_template: A python format string for describing
                how the error reports should look to the user.

        Returns:
            A string representation of the errors.

        """
        return str(self.get_error_report(
            verbosity, filename, message_template
        ))



import ast

from typing import (
    Any,
    Dict,
    List,
)


class AssertVisitor(ast.NodeVisitor):

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None

        # Allow the raise visitor to be used in a mixin.
        # TODO: https://github.com/python/mypy/issues/4001
        super(AssertVisitor, self).__init__(*args, **kwargs)  # type: ignore

        self.asserts = list()  # type: List[ast.Assert]

    def visit_Assert(self, node):
        # type: (ast.Assert) -> ast.AST
        self.asserts.append(node)
        return self.generic_visit(node)
